# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PyTorch optimization for BERT model."""

import math
import torch
from torch.optim import Optimizer
from torch.optim.optimizer import required
from torch.nn.utils import clip_grad_norm_
import logging
import abc
import sys

logger = logging.getLogger(__name__)

if sys.version_info >= (3, 4):
  ABC = abc.ABC
else:
  ABC = abc.ABCMeta('ABC', (), {})


class _LRSchedule(ABC):
  """ Parent of all LRSchedules here. """
  warn_t_total = False  # is set to True for schedules where progressing beyond t_total steps doesn't make sense

  def __init__(self, warmup=0.002, t_total=-1, **kw):
    """
        :param warmup:  what fraction of t_total steps will be used for linear warmup
        :param t_total: how many training steps (updates) are planned
        :param kw:
        """
    super(_LRSchedule, self).__init__(**kw)
    if t_total < 0:
      logger.warning(
        "t_total value of {} results in schedule not being applied".format(
          t_total))
    if not 0.0 <= warmup < 1.0 and not warmup == -1:
      raise ValueError(
        "Invalid warmup: {} - should be in [0.0, 1.0[ or -1".format(warmup))
    warmup = max(warmup, 0.)
    self.warmup, self.t_total = float(warmup), float(t_total)
    self.warned_for_t_total_at_progress = -1

  def get_lr(self, step, nowarn=False):
    """
        :param step:    which of t_total steps we're on
        :param nowarn:  set to True to suppress warning regarding training beyond specified 't_total' steps
        :return:        learning rate multiplier for current update
        """
    if self.t_total < 0:
      return 1.
    progress = float(step) / self.t_total
    ret = self.get_lr_(progress)
    # warning for exceeding t_total (only active with warmup_linear
    if not nowarn and self.warn_t_total and progress > 1. and progress > self.warned_for_t_total_at_progress:
      logger.warning(
        "Training beyond specified 't_total'. Learning rate multiplier set to {}. Please set 't_total' of {} correctly."
        .format(ret, self.__class__.__name__))
      self.warned_for_t_total_at_progress = progress
    # end warning
    return ret

  @abc.abstractmethod
  def get_lr_(self, progress):
    """
        :param progress:    value between 0 and 1 (unless going beyond t_total steps) specifying training progress
        :return:            learning rate multiplier for current update
        """
    return 1.


class ConstantLR(_LRSchedule):
  def get_lr_(self, progress):
    return 1.


class WarmupCosineSchedule(_LRSchedule):
  """
    Linearly increases learning rate from 0 to 1 over `warmup` fraction of training steps.
    Decreases learning rate from 1. to 0. over remaining `1 - warmup` steps following a cosine curve.
    If `cycles` (default=0.5) is different from default, learning rate follows cosine function after warmup.
    """
  warn_t_total = True

  def __init__(self, warmup=0.002, t_total=-1, cycles=.5, **kw):
    """
        :param warmup:      see LRSchedule
        :param t_total:     see LRSchedule
        :param cycles:      number of cycles. Default: 0.5, corresponding to cosine decay from 1. at progress==warmup and 0 at progress==1.
        :param kw:
        """
    super(WarmupCosineSchedule, self).__init__(
      warmup=warmup, t_total=t_total, **kw)
    self.cycles = cycles

  def get_lr_(self, progress):
    if progress < self.warmup:
      return progress / self.warmup
    else:
      progress = (progress - self.warmup) / (1 - self.warmup
                                             )  # progress after warmup
      return 0.5 * (1. + math.cos(math.pi * self.cycles * 2 * progress))


class WarmupCosineWithHardRestartsSchedule(WarmupCosineSchedule):
  """
    Linearly increases learning rate from 0 to 1 over `warmup` fraction of training steps.
    If `cycles` (default=1.) is different from default, learning rate follows `cycles` times a cosine decaying
    learning rate (with hard restarts).
    """

  def __init__(self, warmup=0.002, t_total=-1, cycles=1., **kw):
    super(WarmupCosineWithHardRestartsSchedule, self).__init__(
      warmup=warmup, t_total=t_total, cycles=cycles, **kw)
    assert (cycles >= 1.)

  def get_lr_(self, progress):
    if progress < self.warmup:
      return progress / self.warmup
    else:
      progress = (progress - self.warmup) / (1 - self.warmup
                                             )  # progress after warmup
      ret = 0.5 * (1. + math.cos(math.pi * ((self.cycles * progress) % 1)))
      return ret


class WarmupCosineWithWarmupRestartsSchedule(
    WarmupCosineWithHardRestartsSchedule):
  """
    All training progress is divided in `cycles` (default=1.) parts of equal length.
    Every part follows a schedule with the first `warmup` fraction of the training steps linearly increasing from 0. to 1.,
    followed by a learning rate decreasing from 1. to 0. following a cosine curve.
    """
  warn_t_total = False

  def __init__(self, warmup=0.002, t_total=-1, cycles=1., **kw):
    assert (warmup * cycles < 1.)
    warmup = warmup * cycles if warmup >= 0 else warmup
    super(WarmupCosineWithWarmupRestartsSchedule, self).__init__(
      warmup=warmup, t_total=t_total, cycles=cycles, **kw)

  def get_lr_(self, progress):
    progress = progress * self.cycles % 1.
    if progress < self.warmup:
      return progress / self.warmup
    else:
      progress = (progress - self.warmup) / (1 - self.warmup
                                             )  # progress after warmup
      ret = 0.5 * (1. + math.cos(math.pi * progress))
      return ret


class WarmupConstantSchedule(_LRSchedule):
  """
    Linearly increases learning rate from 0 to 1 over `warmup` fraction of training steps.
    Keeps learning rate equal to 1. after warmup.
    """

  def get_lr_(self, progress):
    if progress < self.warmup:
      return progress / self.warmup
    return 1.


class WarmupLinearSchedule(_LRSchedule):
  """
    Linearly increases learning rate from 0 to 1 over `warmup` fraction of training steps.
    Linearly decreases learning rate from 1. to 0. over remaining `1 - warmup` steps.
    """
  warn_t_total = False

  def get_lr_(self, progress):
    if progress < self.warmup:
      return progress / self.warmup
    return max((progress - 1.) / (self.warmup - 1.), 0.2)


SCHEDULES = {
  None: ConstantLR,
  "none": ConstantLR,
  "warmup_cosine": WarmupCosineSchedule,
  "warmup_constant": WarmupConstantSchedule,
  "warmup_linear": WarmupLinearSchedule,
  "warmup_cosine_warmup_restarts": WarmupCosineWithWarmupRestartsSchedule,
  "warmup_cosine_hard_restarts": WarmupCosineWithHardRestartsSchedule
}


class BertAdam(Optimizer):
  """Implements BERT version of Adam algorithm with weight decay fix.
    Params:
        lr: learning rate
        warmup: portion of t_total for the warmup, -1  means no warmup. Default: -1
        t_total: total number of training steps for the learning
            rate schedule, -1  means constant learning rate of 1. (no warmup regardless of warmup setting). Default: -1
        schedule: schedule to use for the warmup (see above).
            Can be `'warmup_linear'`, `'warmup_constant'`, `'warmup_cosine'`, `'none'`, `None` or a `_LRSchedule` object (see below).
            If `None` or `'none'`, learning rate is always kept constant.
            Default : `'warmup_linear'`
        b1: Adams b1. Default: 0.9
        b2: Adams b2. Default: 0.999
        e: Adams epsilon. Default: 1e-6
        weight_decay: Weight decay. Default: 0.01
        max_grad_norm: Maximum norm for the gradients (-1 means no clipping). Default: 1.0
    """

  def __init__(self,
               params,
               lr=required,
               warmup=-1,
               t_total=-1,
               schedule='warmup_linear',
               b1=0.9,
               b2=0.999,
               e=1e-6,
               weight_decay=0.01,
               max_grad_norm=1.0,
               **kwargs):
    if lr is not required and lr < 0.0:
      raise ValueError(
        "Invalid learning rate: {} - should be >= 0.0".format(lr))
    if not isinstance(schedule, _LRSchedule) and schedule not in SCHEDULES:
      raise ValueError("Invalid schedule parameter: {}".format(schedule))
    if not 0.0 <= b1 < 1.0:
      raise ValueError(
        "Invalid b1 parameter: {} - should be in [0.0, 1.0[".format(b1))
    if not 0.0 <= b2 < 1.0:
      raise ValueError(
        "Invalid b2 parameter: {} - should be in [0.0, 1.0[".format(b2))
    if not e >= 0.0:
      raise ValueError(
        "Invalid epsilon value: {} - should be >= 0.0".format(e))
    # initialize schedule object
    if not isinstance(schedule, _LRSchedule):
      schedule_type = SCHEDULES[schedule]
      schedule = schedule_type(warmup=warmup, t_total=t_total)
    else:
      if warmup != -1 or t_total != -1:
        logger.warning(
          "warmup and t_total on the optimizer are ineffective when _LRSchedule object is provided as schedule. "
          "Please specify custom warmup and t_total in _LRSchedule object.")
    defaults = dict(
      lr=lr,
      schedule=schedule,
      b1=b1,
      b2=b2,
      e=e,
      weight_decay=weight_decay,
      max_grad_norm=max_grad_norm)
    super(BertAdam, self).__init__(params, defaults)

  def get_lr(self):
    lr = []
    for group in self.param_groups:
      for p in group['params']:
        state = self.state[p]
        if len(state) == 0:
          return [0]
        lr_scheduled = group['lr']
        lr_scheduled *= group['schedule'].get_lr(state['step'])
        lr.append(lr_scheduled)
    return lr

  def step(self, closure=None):
    """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
    loss = None
    if closure is not None:
      loss = closure()

    for group in self.param_groups:
      for p in group['params']:
        if p.grad is None:
          continue
        grad = p.grad.data
        if grad.is_sparse:
          raise RuntimeError(
            'Adam does not support sparse gradients, please consider SparseAdam instead'
          )

        state = self.state[p]

        # State initialization
        if len(state) == 0:
          state['step'] = 0
          # Exponential moving average of gradient values
          state['next_m'] = torch.zeros_like(p.data)
          # Exponential moving average of squared gradient values
          state['next_v'] = torch.zeros_like(p.data)

        next_m, next_v = state['next_m'], state['next_v']
        beta1, beta2 = group['b1'], group['b2']

        # Add grad clipping
        if group['max_grad_norm'] > 0:
          clip_grad_norm_(p, group['max_grad_norm'])

        # Decay the first and second moment running average coefficient
        # In-place operations to update the averages at the same time
        next_m.mul_(beta1).add_(1 - beta1, grad)
        next_v.mul_(beta2).addcmul_(1 - beta2, grad, grad)
        update = next_m / (next_v.sqrt() + group['e'])

        # Just adding the square of the weights to the loss function is *not*
        # the correct way of using L2 regularization/weight decay with Adam,
        # since that will interact with the m and v parameters in strange ways.
        #
        # Instead we want to decay the weights in a manner that doesn't interact
        # with the m/v parameters. This is equivalent to adding the square
        # of the weights to the loss with plain (non-momentum) SGD.
        if group['weight_decay'] > 0.0:
          update += group['weight_decay'] * p.data

        lr_scheduled = group['lr']
        lr_scheduled *= group['schedule'].get_lr(state['step'])

        update_with_lr = lr_scheduled * update
        p.data.add_(-update_with_lr)

        state['step'] += 1

        # step_size = lr_scheduled * math.sqrt(bias_correction2) / bias_correction1
        # No bias correction
        # bias_correction1 = 1 - beta1 ** state['step']
        # bias_correction2 = 1 - beta2 ** state['step']

    return loss

class MultipleOptimizer(object):
  def __init__(self, optim, optim_type):
    self.optimizers = optim
    self.min_loss_step = -1
    self.min_loss = 100000
    self.current_step = -1
    self.optim_type = optim_type
    self.stage_two = False
    self.stage_three = False

  def zero_grad(self):
    # if self.optim_type != "two_stage_optim":
    #   for n, op in self.optimizers.items():
    #     op.zero_grad()
    # else:
    #   if self.stage_two:
    #     for n, op in self.optimizers.items():
    #       op.zero_grad()
    #   else:
    #     for n, op in self.optimizers.items():
    #       if n == "bert_optimizer":
    #         continue
    #       op.zero_grad()
      # for n, op in self.optimizers.items():
      #   if n == "bert_optimizer":
      #     op.zero_grad()
      #   else:
      #     op.zero_grad()
    for n, op in self.optimizers.items():
       op.zero_grad()

  def step(self):
    if self.optim_type != "two_stage_optim":
      for n, op in self.optimizers.items():
        op.step()
    else:
      # for n, op in self.optimizers.items():
      #   if n == "bert_optimizer":
      #     op.step()
      #   else:
      #     op.step()
      # self.current_step += 1
      # if self.stage_three:
      #   for n, op in self.optimizers.items():
      #     if n != "ber_optimizer":
      #       continue
      #     op.step()
      if self.stage_two:
        for n, op in self.optimizers.items():
          if n == "bert_optimizer":
            op.step()
        # if self.current_step == 6000:
        #   print("Into Stage Three")
        #   self.stage_three = True
      else:
        for n, op in self.optimizers.items():
          if n == "bert_optimizer":
            continue
          op.step()
        self.current_step += 1
        if self.current_step == 3000:
          print("Into Stage Two")
          self.stage_two = True

        if self.stage_two:
          opti_names = []
          for n in self.optimizers:
            if n != "bert_optimizer":
              opti_names.append(n)
          for n in opti_names:
            optimizer = self.optimizers.pop(n)
            del optimizer
            torch.cuda.empty_cache()

  def update_loss(self, loss):

    # if loss < self.min_loss:
    #   self.min_loss = loss
    #   self.min_loss_step = self.current_step
    # if not self.stage_two and self.optim_type == "two_stage_optim" and self.current_step > self.min_loss_step + 5:
    #   self.stage_two = True
    #   print("Into stage two")
    if self.current_step == 2000:
      self.stage_two = True



def adjust_learning_rate(optimizer, config, epoch):
  """Sets the learning rate to the initial LR decayed by 10 every 30 epochs"""
  lr = config.sgd_learning_rate * (0.8 ** (epoch // 1))
  for param_group in optimizer.param_groups:
    param_group['lr'] = lr


import sys
import os
try:
    import cPickle as pickle
except ImportError:
    import pickle


class Memoize(object):
  def __init__(self, f):
    self.f = f
    self.cache = {}

  def __call__(self, *args):
    if args not in self.cache:
      self.cache[args] = self.f(*args)
    return self.cache[args]

def load_pickle(path, memoized=True):
  return _load_pickle_memoize(path) if memoized else _load_pickle(path)

def _load_pickle(path):
  with open(path, 'rb') as f:
    return pickle.load(f)

@Memoize
def _load_pickle_memoize(path):
  return _load_pickle(path)


def write_pickle(o, path):
  dir = path.rsplit('/', 1)[0]
  if not os.path.exists(dir):
    os.mkdir(dir)
  with open(path, 'wb') as f:
    pickle.dump(o, f, -1)

def log(*args):
  msg = ' '.join(map(str, args))
  sys.stdout.write(msg + '\n')
  sys.stdout.flush()


def heading(*args):
  log()
  log(80 * '=')
  log(*args)
  log(80 * '=')


from relogic.logickit.base.constants import (JOINT_SRL_TASK, PIPE_SRL_TASK, ECP_TASK, POINTWISE_TASK, IR_TASK, NER_TASK,
  SEQUENCE_LABELING_TASK, PARALLEL_MAPPING_TASK, PAIRWISE_TASK, PARALLEL_TEACHER_STUDENT_TASK,
  SEQUENCE_CLASSIFICATION_TASK, ENTITY_TYPE_CLASSIFICATION, DEP_PARSING_TASK, MIXSENT_TASK, LANGUAGE_IDENTIFICATION_IR,
  POS_TASK, DOCIR_TASK, LANGUAGE_IDENTIFICATION_SEQ, PREDICATE_DETECTION_TASK, IR_SIAMESE_TASK, IR_SAMCNN_TASK)
from relogic.logickit.dataflow.dataflow import DataFlow, Example, Feature, MiniBatch
from relogic.logickit.dataflow.srl import SRLDataFlow, SRLExample, SRLFeature, SRLMiniBatch
from relogic.logickit.dataflow.ecp import ECPDataFlow, ECPExample, ECPFeature, ECPMiniBatch
from relogic.logickit.dataflow.pointwise import PointwiseDataFlow, PointwiseExample, PointwiseFeature, PointwiseMiniBatch
from relogic.logickit.dataflow.sequence import SequenceDataFlow, SequenceExample, SequenceFeature, SequenceMiniBatch
from relogic.logickit.dataflow.parallel import ParallelDataFlow, ParallelExample, ParallelFeature, ParallelMiniBatch
from relogic.logickit.dataflow.pairwise import PairwiseDataFlow, PairwiseExample, PairwiseFeature, PairwiseMiniBatch
from relogic.logickit.dataflow.singleton import SingletonDataFlow, SingletonExample, SingletonFeature, SingletonMiniBatch
from relogic.logickit.dataflow.dep import (DependencyParsingDataFlow,
  DependencyParsingExample, DependencyParsingFeature, DependencyParsingMiniBatch)
from relogic.logickit.dataflow.mixsent import MixSentDataFlow, MixSentExample, MixSentFeature, MixSentMiniBatch
from relogic.logickit.dataflow.doc_pointwise import (DocPointwiseDataFlow,
  DocPointwiseExample, DocPointwiseFeature, DocPointwiseMiniBatch)
from relogic.logickit.dataflow.twins import TwinsDataFlow, TwinsExample, TwinsFeature, TwinsMiniBatch
from relogic.common.prefix_map import PrefixMap

task_to_dataflow_class_map = {
  PIPE_SRL_TASK: SRLDataFlow,
  JOINT_SRL_TASK: SRLDataFlow,
  ECP_TASK: ECPDataFlow,
  POINTWISE_TASK: PointwiseDataFlow,
  IR_TASK: PointwiseDataFlow,
  IR_SIAMESE_TASK: TwinsDataFlow,
  IR_SAMCNN_TASK: TwinsDataFlow,
  SEQUENCE_LABELING_TASK: SequenceDataFlow,
  POS_TASK: SequenceDataFlow,
  NER_TASK: SequenceDataFlow,
  PREDICATE_DETECTION_TASK: SequenceDataFlow,
  PARALLEL_MAPPING_TASK: ParallelDataFlow,
  PAIRWISE_TASK: PairwiseDataFlow,
  PARALLEL_TEACHER_STUDENT_TASK: ParallelDataFlow,
  SEQUENCE_CLASSIFICATION_TASK: SingletonDataFlow,
  ENTITY_TYPE_CLASSIFICATION: SingletonDataFlow,
  DEP_PARSING_TASK: DependencyParsingDataFlow,
  MIXSENT_TASK: MixSentDataFlow,
  LANGUAGE_IDENTIFICATION_IR: PointwiseDataFlow,
  LANGUAGE_IDENTIFICATION_SEQ: SingletonDataFlow,
  DOCIR_TASK: DocPointwiseDataFlow
}

TASK_TO_DATAFLOW_CLASS_MAP = PrefixMap(task_to_dataflow_class_map)


import torch.nn as nn
import torch
from relogic.logickit.modules.adversarial.discriminator import LR
import numpy as np
from relogic.logickit.base.configuration import AdversarialConfigs


class Adversarial(nn.Module):
  def __init__(self, config: AdversarialConfigs):
    super().__init__()
    self.config = config

    if self.config.discriminator_type == "LR":
      self.discriminator = LR(config=config)
    else:
      NotImplementedError()

    print("Using Adversarial {}".format(config.type))

    if config.type == "WGAN":
      pass

    if config.type == "GAN" or config.type == "GR":
      # Let make it adam for now.
      self.optim = torch.optim.Adam(self.discriminator.parameters(),
                                    lr=self.config.discriminator_lr)



    self.real = 1
    self.fake = 0
    if config.soft_label and config.type == "GAN":
      self.real = np.random.uniform(0.7, 1.0)
      self.fake = np.random.uniform(0.0, 0.3)
      print("The soft label is {} and {}".format(self.real, self.fake))

    if config.type == "GR":
      self.criterion = nn.CrossEntropyLoss()
    else:
      self.criterion = nn.BCEWithLogitsLoss()

  def loss(self, input, label):
    output = self.discriminator(features=input)
    assert (output.dim() == 2)

    if self.config.type == "WGAN":
      loss = torch.mean(output)
    else:
      if self.config.type == "GAN":
        label = torch.empty(*output.size()).fill_(label).type_as(output)
      elif self.config.type == "GR":
        label = torch.empty(output.size(0)).fill_(label).type_as(output).long()

      loss = self.criterion(output, label)
    return output, loss

  def update(self, real_in, fake_in, real_id, fake_id):
    self.optim.zero_grad()

    if self.config.type == "GAN":
      real_id, fake_id = self.real, self.fake

    real_output, real_loss = self.loss(real_in, real_id)
    fake_output, fake_loss = self.loss(fake_in, fake_id)

    if self.config.type in ["GR", "GAN"]:
      loss = 0.5 * (real_loss + fake_loss)
    else:
      loss = fake_loss - real_loss
    loss.backward()
    self.optim.step()

    real_acc, fake_acc = 0, 0
    if self.config.type in ["GR", "GAN"]:
      real_acc = self.accuracy(real_output, real_id)
      fake_acc = self.accuracy(fake_output, fake_id)

    return real_loss.item(), fake_loss.item(), real_acc, fake_acc

  def accuracy(self, output, label):
    if self.config.type == "GAN":
      if label > 0.5:
        preds = (torch.sigmoid(output) >= 0.5).long().cpu()
      else:
        preds = (torch.sigmoid(output) < 0.5).long().cpu()
      label = 1

    labels = torch.LongTensor([label])
    labels = labels.expand(*preds.size())
    n_correct = preds.eq(labels).sum().item()
    acc = 1.0 * n_correct / output.size(0)
    return acc

  def gen_loss(self, real_in, fake_in, real_id, fake_id):
    """Functions to calculate loss to update the Generator.
       The basic idea is to minimize the loss that can confuse the Discriminator
       to make a wrong decision."""
    if self.config.type == "GAN":
      _, loss = self.loss(fake_in, self.real)
      loss = self.config.scale * loss
    elif self.config.type == "WGAN":
      # clamp parameters to a cube
      for p in self.discriminator.parameters():
        p.data.clamp_(self.config.clip_lower, self.config.clip_upper)
      _, loss = self.loss(fake_in, self.real)
      loss = -self.config.scale * loss
    else:
      raise NotImplementedError()
    return loss

  def forward(self, *input, **kwargs):
    raise NotImplementedError()





from relogic.logickit.scorer.tagging_scorers import F1Score, softmax
from relogic.logickit.utils.utils import get_span_labels
from relogic.logickit.scorer.span_level_scorer import SpanLevelScorer
from relogic.logickit.scorer.scorer import Scorer
import os
import torch
import json
from relogic.logickit.dataflow.srl import SRLExample

class SpanSRLF1Scorer(SpanLevelScorer):
  def __init__(self, label_mapping, dump_to_file=None):
    super(SpanSRLF1Scorer, self).__init__()
    self._inv_label_mapping = {v:k for k, v in label_mapping.items()}
    if dump_to_file:
      self.dump_to_file_path = os.path.join(dump_to_file["output_dir"], dump_to_file["task_name"] + "_dump.json")
      self.dump_to_file_handler = open(self.dump_to_file_path, 'w')

  def _get_results(self):
    self._n_correct, self._n_predicted, self._n_gold = 0, 0, 0
    for example, boundary, pred_label in zip(self._examples, self._boundaries, self._preds):
      pred_tags = pred_label.argmax(-1).data.cpu().numpy()

      sent_spans, sent_labels = get_span_labels(
        sentence_tags=example.label)

      # sent_spans = set([(b[0], b[1], l) for b, l in zip(example.spans, example.span_labels)])
      span_preds = set([(int(b[0]), int(b[1]), self._inv_label_mapping[l]) for b, l in zip(example.span_candidates, pred_tags) if b[0] >= 0])
      span_preds = set(filter(lambda item: item[0] != example.predicate_index, span_preds))
      self._n_correct += len(sent_spans & span_preds)
      self._n_gold += len(sent_spans)
      self._n_predicted += len(span_preds)
      if self.dump_to_file_path:
        self.dump_to_file_handler.write(json.dumps({
          "text": example.raw_text,
          "predicate_text": example.predicate_text,
          "predicate_index": example.predicate_index,
          "span_preds": list(span_preds),
          "sent_spans": list(sent_spans)}) + "\n")
    if self.dump_to_file_path:
      self.dump_to_file_handler.close()
    # TODO: a quick fix for calculating the F1 score
    if self._n_correct == 0:
      p, r, f1 = 0, 0, 0
    else:
      p = 100.0 * self._n_correct / self._n_predicted
      r = 100.0 * self._n_correct / self._n_gold
      f1 = 2 * p * r / (p + r)
    return [
      ("total entity", self._n_gold),
      ("find entity", self._n_predicted),
      ("correct entity", self._n_correct),
      ("precision", p),
      ("recall", r),
      ("f1", f1),
      ("loss", self.get_loss()),
    ]
  def get_loss(self):
    return 0

class SRLF1Scorer(F1Score):
  def __init__(self, label_mapping, dump_to_file=None):
    super(SRLF1Scorer, self).__init__()
    # label mapping : str -> int
    self._inv_label_mapping = {v: k for k, v in label_mapping.items()}
    if dump_to_file:
      self.dump_to_file_path = os.path.join(dump_to_file["output_dir"], dump_to_file["task_name"] + "_dump.json")
      self.dump_to_file_handler = open(self.dump_to_file_path, 'w')

  def _get_results(self):
    # The scorer will be cleared after the evaluation is done. 

    self._n_correct, self._n_predicted, self._n_gold = 0, 0, 0
    for example, preds in zip(self._examples, self._preds):
      example: SRLExample
      preds_tags = preds.argmax(-1).data.cpu().numpy()
      confidences = [max(softmax(token_level)) for token_level in preds.data.cpu().numpy()]
      sent_spans, sent_labels = get_span_labels(
        sentence_tags = example.labels,
        ignore_label=["V"]
        )
      sent_spans = set(filter(lambda item: item[0] != example.predicate_index, sent_spans))
      span_preds, pred_labels = get_span_labels(
        sentence_tags=preds_tags,
        is_head = example.is_head,
        segment_id = example.segment_ids,
        inv_label_mapping = self._inv_label_mapping,
        ignore_label=["V"])
      span_preds = set(filter(lambda item: item[0] != example.predicate_index, span_preds))
      self._n_correct += len(sent_spans & span_preds)
      self._n_gold += len(sent_spans)
      self._n_predicted += len(span_preds)
      if self.dump_to_file_path:
        if len(example.raw_tokens) != len(sent_labels) or len(example.raw_tokens) != len(pred_labels):
          print(len(example.raw_tokens), example.raw_tokens)
          print(len(sent_labels), sent_labels)
          print(len(pred_labels), pred_labels)
          exit()
        self.dump_to_file_handler.write(json.dumps({
          "text": example.raw_tokens,
          "predicate_text": example.predicate_text,
          "predicate_index": example.predicate_index,
          "label": sent_labels,
          "predicted": pred_labels
        }) + "\n")
        # data_to_dump.append(((int(example.guid), int(example.predicate_index)), example.raw_text, sent_labels, pred_labels, confidences))

    if self.dump_to_file_path:
      self.dump_to_file_handler.close()

    return super(SRLF1Scorer, self)._get_results()

class JointSpanSRLF1Scorer(Scorer):
  def __init__(self, label_mapping, dump_to_file=None):
    super(JointSpanSRLF1Scorer, self).__init__()
    self._inv_label_mapping = {v: k for k, v in label_mapping.items()}
    self._examples = []
    self._predictions = []
    self._n_correct = 0
    self._n_gold = 0
    self._n_predicted = 0
    if dump_to_file:
      self.dump_to_file_path = os.path.join(dump_to_file["output_dir"], dump_to_file["task_name"] + "_dump.json")
      self.dump_to_file_handler = open(self.dump_to_file_path, 'w')

  def update(self, mbs, predictions, loss, extra_args):
    # predictions = srl_scores, top_pred_spans, top_arg_spans
    srl_scores, top_pred_spans, top_arg_spans, top_pred_span_mask, top_arg_span_mask, _, _, _ = predictions
    # top_pred_spans = (batch, max_seq_len, 2)
    batch_size = top_pred_spans.size(0)

    max_pred_num = top_pred_spans.size(1)
    max_arg_num = top_arg_spans.size(1)
    expanded_top_pred_spans = top_pred_spans.unsqueeze(2).repeat(1, 1, max_arg_num, 1)
    expanded_top_arg_spans = top_arg_spans.unsqueeze(1).repeat(1, max_pred_num, 1, 1)
    expanded_top_pred_span_mask = top_pred_span_mask.unsqueeze(2).repeat(1, 1, max_arg_num)
    expanded_top_arg_span_mask = top_arg_span_mask.unsqueeze(1).repeat(1, max_pred_num, 1)
    indices_mask = expanded_top_pred_span_mask & expanded_top_arg_span_mask

    indices = torch.cat([expanded_top_pred_spans, expanded_top_arg_spans], dim=-1)
    # batch_id = torch.arange(0, batch_size).unsqueeze(1).unsqueeze(1).unsqueeze(1).repeat(1, *indices.size()[1:3], 1).to(
    #   indices.device)
    # expanded_indices = torch.cat([batch_id, indices], dim=-1)
    # flatten_expanded_indices = expanded_indices.view(-1, 5)

    preds_tags = srl_scores.argmax(-1)
    if len(expanded_top_pred_spans) != len(expanded_top_arg_spans) or len(preds_tags) != len(expanded_top_arg_spans):
      raise ValueError("The length of predicate span {} is not equal to argument span {}".format(
        len(expanded_top_pred_spans), len(expanded_top_arg_spans)))

    for example_indices, example_pred, example_indices_mask in zip(indices, preds_tags, indices_mask):
      prediction_list = []
      for indices, pred, mask in zip(example_indices.view(-1, 4), example_pred.view(-1), example_indices_mask.view(-1)):
        if mask:
          pred_label = self._inv_label_mapping[pred.data.cpu().numpy().item()]
          # if pred_label != 'X' and pred_label != 'O':
          prediction_list.append(tuple(indices.data.cpu().numpy()) +  (pred_label,))
      self._predictions.append(prediction_list)
    for example in mbs.examples:
      self._examples.append(example)



  def _get_results(self):
    assert len(self._examples) == len(self._predictions)
    for example, preds in zip(self._examples, self._predictions):
      # sent_spans = set([(pred_start, pred_end, arg_start, arg_end, self._inv_label_mapping[label_id]) for
      #               pred_start, pred_end, arg_start, arg_end, label_id in example.label_ids])
      # TODO: fix for LSTM
      sent_spans = set([(pred_start, pred_end, arg_start, arg_end, self._inv_label_mapping[label_id]) for
                        pred_start, pred_end, arg_start, arg_end, label_id in example.label_ids])
      example: SRLExample
      without_filter_span_preds = set(preds)
      sent_spans = set(filter(lambda item: item[4] != 'V', sent_spans))
      span_preds = set(filter(lambda item: item[4] not in ['V', "O", "X"], without_filter_span_preds))
      self._n_correct += len(sent_spans & span_preds)
      self._n_gold += len(sent_spans)
      self._n_predicted += len(span_preds)

      if self.dump_to_file_path:
        sent_spans = [(int(s[0]), int(s[1]), int(s[2]), int(s[3]), s[4]) for s in sent_spans]
        span_preds = [(int(s[0]), int(s[1]), int(s[2]), int(s[3]), s[4]) for s in span_preds]
        without_filter_span_preds = [(int(s[0]), int(s[1]), int(s[2]), int(s[3]), s[4]) for s in without_filter_span_preds]
        self.dump_to_file_handler.write(json.dumps({
          "text": example.raw_tokens,
          "BPE_tokens": example.text_tokens,
          "span_preds": list(span_preds),
          "sent_spans": list(sent_spans),
          "pruned_span_preds": list(without_filter_span_preds)}) + "\n")
    if self.dump_to_file_path:
      self.dump_to_file_handler.close()

    if self._n_correct == 0:
      p, r, f1 = 0, 0, 0
    else:
      p = 100.0 * self._n_correct / self._n_predicted
      r = 100.0 * self._n_correct / self._n_gold
      f1 = 2 * p * r / (p + r)
    return [
      ("total entity", self._n_gold),
      ("find entity", self._n_predicted),
      ("correct entity", self._n_correct),
      ("precision", p),
      ("recall", r),
      ("f1", f1),
      ("loss", self.get_loss()),
    ]

  def get_loss(self):
    return 0


import torch
import torch.nn as nn

class ClassificationModule(nn.Module):
  def __init__(self, config, task_name, n_classes):
    super(ClassificationModule, self).__init__()
    self.config = config
    self.task_name = task_name
    self.n_classes = n_classes
    self.to_logits = nn.Linear(config.hidden_size, self.n_classes)

  def forward(self, *inputs, **kwargs):
    features = kwargs.pop("features")
    if isinstance(features, list):
      logits = self.to_logits(features[0][:, 0])
    else:
      logits = self.to_logits(features[:, 0])
    return logits


from relogic.components.component import Component
from relogic.structures.structure import Structure
from typing import List


class SRLComponent(Component):
  """

  """

  def execute(self, inputs: List[Structure]):
    counter = 0
    expanded_inputs = []
    mapping = []
    for idx, structure in enumerate(inputs):
      for predicate_index, predicate_text in structure.predicates:
        expanded_inputs.append(
          structure.__class__(
            tokens=structure.tokens, predicate_text=predicate_text, predicate_index=predicate_index))
        mapping.append(idx)
    for results in self._predictor.predict(expanded_inputs):
      _, batch_labels, _ = results
      for labels in batch_labels:
        inputs[mapping[counter]].srl_labels.append(list(labels))
        counter += 1


from relogic.components.component import Component
from relogic.structures.structure import Structure
from typing import List


class PredicateDetectionComponent(Component):
  """

  """


  @classmethod
  def from_pretrained(cls, pretrained_model_name_or_path, *inputs, **kwargs):
    import spacy

    if pretrained_model_name_or_path == "spacy":
      return PredicateDetectionComponent(model_name=pretrained_model_name_or_path, config=None,
                                         predictor=spacy.load("en_core_web_sm", disable=["parser", "ner"]))
    else:
      return super().from_pretrained(pretrained_model_name_or_path=pretrained_model_name_or_path)

  def remerge_sent(self, sent):
    i = 0
    while i < len(sent) - 1:
      tok = sent[i]
      if not tok.whitespace_:
        ntok = sent[i + 1]
        # in-place operation.
        sent.merge(tok.idx, ntok.idx + len(ntok))
      i += 1
    return sent

  def execute(self, inputs: List[Structure]):
    if self.model_name == "spacy":
      for structure in inputs:
        results = self.remerge_sent(self._predictor(structure.tokenized_text))
        for idx, token in enumerate(results):
          structure.pos.append(token.pos_)
          if token.pos_ == "VERB":
            structure.predicates.append((idx, token.text))
    else:
      counter = 0
      for results in self._predictor.predict(inputs):
        _, batch_labels, _ = results
        for labels in batch_labels:
          for idx, label in enumerate(labels):
            if label == "B-V":
              inputs[counter].predicates. append((idx, inputs[counter][idx].text))
          counter += 1



import torch.nn as nn
import torch
from relogic.logickit.utils.utils import batched_index_select_tensor

class MixSentAlignmentModule(nn.Module):
  def __init__(self, config, task_name, n_classes=None):
    super().__init__()
    self.config = config
    self.task_name = task_name

  def forward(self, *input, **kwargs):
    student_results = kwargs.pop("student_results")
    teacher_results = kwargs.pop("teacher_results")

    logits_a = teacher_results["a"]
    logits_b = teacher_results["b"]
    logits_c = student_results

    span_a_selected_index = kwargs.pop("span_a_selected_index")
    span_b_selected_index = kwargs.pop("span_b_selected_index")
    span_c_a_selected_index = kwargs.pop("span_c_a_selected_index")
    span_c_b_selected_index = kwargs.pop("span_c_b_selected_index")
    # The selected index should be all greater than 0

    selected_logits_a = batched_index_select_tensor(logits_a, span_a_selected_index)
    selected_logits_b = batched_index_select_tensor(logits_b, span_b_selected_index)
    selected_logits_c_a = batched_index_select_tensor(logits_c, span_c_a_selected_index)
    selected_logits_c_b = batched_index_select_tensor(logits_c, span_c_b_selected_index)

    combined_c_teacher = torch.cat([selected_logits_a, selected_logits_b], dim=1)
    combined_c_student = torch.cat([selected_logits_c_a, selected_logits_c_b], dim=1)

    return combined_c_student, combined_c_teacher, None


from relogic.logickit.scorer.scorer import Scorer
from relogic.logickit.utils.utils import softmax, sigmoid
import torch.nn.functional as F
import torch
from tqdm import tqdm
import os
import subprocess
import json

class RecallScorer(Scorer):
  def __init__(self, label_mapping, topk, correct_label='1', dump_to_file=None):
    super(RecallScorer, self).__init__()
    self.label_mapping = label_mapping
    self._inv_label_mapping = {v: k for k, v in label_mapping.items()}
    self._examples = []
    self._preds = []
    self.topk = topk
    self.correct_label = correct_label
    if dump_to_file:
      self.dump_to_file_path = os.path.join(dump_to_file["output_dir"], dump_to_file["task_name"] + "_dump.json")


  def update(self, mbs, predictions, loss, extra_args):
    super(RecallScorer, self).update(mbs, predictions, loss, extra_args)
    for example, preds in zip(mbs.examples, predictions):
      self._examples.append(example)
      self._preds.append(preds)

  def get_loss(self):
    return 0

  def _get_results(self):
    if self.dump_to_file_path:
      self.dump_to_file_handler = open(self.dump_to_file_path, 'w')

    self._n_hit_left, self._n_hit_right, self._n_total_left, self._n_total_right = 0, 0, 0, 0
    pred_collection = [{}, {}] # forward direction and backward direction
    gold_collection = [{}, {}]
    for example, preds in zip(self._examples, self._preds):
      prob = preds[self.label_mapping[self.correct_label]].item()
      query_id, candidate_id, direction = example.guid.split('-')
      direction = int(direction)
      if direction == 1:
        query_id, candidate_id = candidate_id, query_id
      if self.dump_to_file_handler:
        self.dump_to_file_handler.write("{} {} {} {}\n".format(query_id, candidate_id, direction, prob))
      if query_id not in pred_collection[direction]:
        pred_collection[direction][query_id] = []
      pred_collection[direction][query_id].append((candidate_id, prob))

      # if example.label == self.correct_label:
      #   gold_collection[direction][query_id] = candidate_id
      gold_query_id, gold_candidate_id, gold_direction = example.gold_pair.split('-')
      gold_direction = int(gold_direction)
      if gold_direction == 1:
        gold_query_id, gold_candidate_id = gold_candidate_id, gold_query_id
      gold_collection[gold_direction][gold_query_id] = gold_candidate_id

    if len(pred_collection[0]) != len(gold_collection[0]) or len(pred_collection[1]) != len(gold_collection[1]):
      raise ValueError("The query size in pred collectdion {}|{} is different from gold collection {}|{}".format(
        len(pred_collection[0]), len(pred_collection[1]), len(gold_collection[0]), len(gold_collection[1])))
    for d in range(2):
      for query_id in pred_collection[d]:
        sorted_list = sorted(pred_collection[d][query_id], key=lambda x: x[1], reverse=True)
        candidate_ids = [x[0] for x in sorted_list][:self.topk]
        if gold_collection[d][query_id] in candidate_ids:
          if d == 0:
            self._n_hit_left += 1
          else:
            self._n_hit_right += 1
    self._n_total_left = len(gold_collection[0])
    self._n_total_right = len(gold_collection[1])

    if self.dump_to_file_path:
      self.dump_to_file_handler.close()

    return [("hits_left", self._n_hit_left),
            ("hits_right", self._n_hit_right),
            ("total_left", self._n_total_left),
            ("total_right", self._n_total_right),
            ("recall_left", self._n_hit_left / self._n_total_left),
            ("recall_right", self._n_hit_right / self._n_total_right)]

class CartesianMatchingRecallScorer(Scorer):
  def __init__(self, topk, qrels_file_path, dump_to_file=None):
    super(CartesianMatchingRecallScorer, self).__init__()
    self.topk = topk
    self.qrels_file_path = qrels_file_path
    if dump_to_file:
      self.dump_to_file_path = os.path.join(dump_to_file["output_dir"], dump_to_file["task_name"] + "_dump.json")
      self.dump_to_file_handler = open(self.dump_to_file_path, 'w')
    else:
      print("You need to specify the dump_to_file path for the pair matching task")
      exit()

  def update(self, mbs, reprs, loss, extra_args):
    super(CartesianMatchingRecallScorer, self).update(mbs, reprs, loss, extra_args)
    for example, repr in zip(mbs.examples, reprs):
      self.dump_to_file_handler.write(json.dumps({
        "guid": example.guid,
        "feature": " ".join([str(f) for f in repr.cpu().data.numpy()])}) + "\n")

  def get_loss(self):
    return 0

  def _get_results(self):
    self.dump_to_file_handler.close()
    dir = os.path.abspath(os.path.dirname(__file__))
    recall_eval_path = os.path.join(dir, '..', '..', '..', 'evals', 'pair_matching', 'entity_align_eval.py')
    eval_out = subprocess.check_output(["python", recall_eval_path, "-e", self.dump_to_file_path, "-g", self.qrels_file_path])
    eval_out_lines = str(eval_out, 'utf-8').split('\n')
    results = []
    for line in eval_out_lines:
      if line.startswith("Hits@1:"):
        score = float(line.strip().split(" ")[1].strip("%"))
        results.append(score)

    return [("recall_left", results[0]),
            ("recall_right", results[1])]


# class CartesianMatchingRecallScorer(Scorer):
#   def __init__(self, topk, dump_to_file=None):
#     super(CartesianMatchingRecallScorer, self).__init__()
#     self.topk = topk
#     self.left = {}
#     self.right = {}
#     self.left_to_right_gold = {}
#     self.right_to_left_gold = {}
#     if dump_to_file:
#       self.dump_to_file_path = os.path.join(dump_to_file["output_dir"], dump_to_file["task_name"] + "_dump.json")
#
#
#   def update(self, mbs, reprs, loss, extra_args):
#     super(CartesianMatchingRecallScorer, self).update(mbs, reprs, loss, extra_args)
#     for example, repr in zip(mbs.examples, reprs):
#       left_id, right_id, direc = example.guid.split('-')
#       if direc == "0":
#         self.left[left_id] = repr
#       else:
#         self.right[right_id] = repr
#       left_id, right_id, direc = example.gold_pair.split('-')
#       if direc == "0":
#         self.left_to_right_gold[left_id] = right_id
#       else:
#         self.right_to_left_gold[right_id] = left_id
#
#   def get_loss(self):
#     return 0
#
#   def _get_results(self):
#     if self.dump_to_file_path:
#       self.dump_to_file_handler = open(self.dump_to_file_path, 'w')
#
#     self._n_hit_left, self._n_hit_right, self._n_total_left, self._n_total_right = 0, 0, 0, 0
#     if self.dump_to_file_path:
#       for key in self.left:
#         self.dump_to_file_handler.write("{}\t".format(key) + " ".join([str(f) for f in self.left[key].cpu().data.numpy()]) + '\n')
#       for key in self.right:
#         self.dump_to_file_handler.write("{}\t".format(key) + " ".join([str(f) for f in self.right[key].cpu().data.numpy()]) + '\n')
#     left_to_right_distance = {}
#     right_to_left_distance = {}
#     print("Evaluating left to right")
#     for left_id in self.left.keys():
#       left_to_right_distance[left_id] = {}
#       for right_id in self.right.keys():
#         left_to_right_distance[left_id][right_id] = torch.sum(torch.abs(self.left[left_id] - self.right[right_id])).item()
#     print("Evaluating right to left")
#     for right_id in self.right.keys():
#       right_to_left_distance[right_id] = {}
#       for left_id in self.left.keys():
#         right_to_left_distance[right_id][left_id] = torch.sum(torch.abs(self.right[right_id] - self.left[left_id])).item()
#
#     for left_id in self.left.keys():
#       sorted_list = sorted(left_to_right_distance[left_id].items(), key=lambda x: x[1])
#       candidate_ids = [x[0] for x in sorted_list][:self.topk]
#       if self.left_to_right_gold[left_id] in candidate_ids:
#         self._n_hit_left += 1
#       # if dump_to_file:
#       #   for right_id, dist in sorted_list:
#       #     fout.write("{}\t{}\t{}\n".format(left_id, right_id, dist))
#     for right_id in self.right.keys():
#       sorted_list = sorted(right_to_left_distance[right_id].items(), key=lambda x: x[1])
#       candidate_ids = [x[0] for x in sorted_list][:self.topk]
#       if self.right_to_left_gold[right_id] in candidate_ids:
#         self._n_hit_right += 1
#       # if dump_to_file:
#       #   for left_id, dist in sorted_list:
#       #     fout.write("{}\t{}\t{}\n".format(right_id, left_id, dist))
#
#     if self.dump_to_file_path:
#       self.dump_to_file_handler.close()
#
#     self._n_total_left = len(self.left_to_right_gold)
#     self._n_total_right = len(self.right_to_left_gold)
#     return [("hits_left", self._n_hit_left),
#             ("hits_right", self._n_hit_right),
#             ("total_left", self._n_total_left),
#             ("total_right", self._n_total_right),
#             ("recall_left", self._n_hit_left / self._n_total_left),
#             ("recall_right", self._n_hit_right / self._n_total_right)]

class RetrievalScorer(Scorer):
  """

  """
  def __init__(self, label_mapping, qrels_file_path, correct_label='1', dump_to_file=None, regression=False):
    super(RetrievalScorer, self).__init__()
    self.label_mapping = label_mapping
    self._inv_label_mapping = {v: k for k, v in label_mapping.items()}
    self._examples = []
    self._preds = []
    self.correct_label = correct_label
    self.qrels_file_path = qrels_file_path
    self.regression = regression
    # Because we need to leverage trec_eval to calculate the scores, so dump_to_file can not be None
    if dump_to_file:
      self.dump_to_file_path = os.path.join(dump_to_file["output_dir"], dump_to_file["task_name"] + "_dump.json")
      self.dump_to_file_handler = open(self.dump_to_file_path, 'w')
    else:
      print("You need to specify the dump_to_file path for the retrieval task")
      exit()

  def update(self, mbs, predictions, loss, extra):
    super(RetrievalScorer, self).update(mbs, predictions, loss, extra)
    # TODO: we are going to migrate the interface !
    predictions = predictions["logits"]
    for example, preds in zip(mbs.examples, predictions):
      self._examples.append(example)
      self._preds.append(preds.data.cpu().numpy())

  def get_loss(self):
    return 0

  def _get_results(self):
    topic_doc_collection = {}
    for example, preds in zip(self._examples, self._preds):
      if self.regression:
        score = sigmoid(preds)[0]
      else:
        preds = softmax(preds)
        score = preds[self.label_mapping[self.correct_label]]
      text_a_id, text_b_id = example.guid.split('|')
      if text_a_id not in topic_doc_collection:
        topic_doc_collection[text_a_id] = {}
      topic_doc_collection[text_a_id][text_b_id] = max(topic_doc_collection[text_a_id].get(text_b_id, 0), score)
    for text_a_id in topic_doc_collection:
      for text_b_id in topic_doc_collection[text_a_id]:
        score = topic_doc_collection[text_a_id][text_b_id]
        self.dump_to_file_handler.write("{} Q0 {} 0 {} rerank\n".format(text_a_id, text_b_id, score))

    self.dump_to_file_handler.flush()
    self.dump_to_file_handler.close()
    dir = os.path.abspath(os.path.dirname(__file__))
    trec_eval_path = os.path.join(dir, '..', '..', '..', 'evals', 'trec_eval', 'trec_eval.9.0.4/trec_eval')

    trec_out = subprocess.check_output([trec_eval_path, self.qrels_file_path , self.dump_to_file_path])
    trec_out_lines = str(trec_out, 'utf-8').split('\n')
    mean_average_precision = float(trec_out_lines[5].split('\t')[-1])
    # mean_reciprocal_rank = float(trec_out_lines[9].split('\t')[-1])
    # p_30 = float(trec_out_lines[25].split('\t')[-1])

    return [("map", mean_average_precision)]






import torch
import torch.nn as nn
from torch.optim import SGD, Adam

from relogic.logickit.base import utils
from relogic.logickit.model.base_model import BaseModel
from relogic.logickit.model.optimization import BertAdam, MultipleOptimizer
import numpy as np
from relogic.logickit.inference import get_inference
from relogic.logickit.dataflow import MiniBatch
from relogic.logickit.utils.utils import entropy
from relogic.logickit.base.configuration import Configuration
from relogic.logickit.inference.adversarial import Adversarial


class Model(BaseModel):
  def __init__(self, config, tasks, ext_config: Configuration=None):
    super(Model, self).__init__(config, ext_config)
    self.tasks = tasks
    utils.log("Building model")
    inference = get_inference(config)(config, tasks)
    utils.log("Switch Model to device")
    inference = inference.to(self.device)
    # TODO: need to test
    if config.multi_gpu:
      inference = torch.nn.DataParallel(inference)
    self.model = inference
    self.teacher = inference
    utils.log(self.model.__str__())

    if config.mode == "train" or config.mode == "finetune":
      self.setup_training(config, tasks)

    ## Inplace Relu
    def inplace_relu(m):
      classname = m.__class__.__name__
      if classname.find('ReLU') != -1:
        m.inplace = True

    inference.apply(inplace_relu)

    if config.adversarial_training:
      self.adversarial_agent = Adversarial(config=ext_config.adversarial_configs).to(self.device)


  def setup_training(self, config, tasks):
    # Calculate optimization steps
    size_train_examples = 0
    config.num_steps_in_one_epoch = 0
    if config.mode == "train" or config.mode == "finetune":
      for task in tasks:
        utils.log("{} : {}  training examples".format(task.name, task.train_set.size))
        if "loss_weight" in config.tasks[task.name]:
          utils.log("loss weight {}".format(config.tasks[task.name]["loss_weight"]))
        size_train_examples += task.train_set.size
        config.num_steps_in_one_epoch += task.train_set.size // config.tasks[task.name]["train_batch_size"]

        # config.train_batch_size = config.train_batch_size // config.gradient_accumulation_steps
        # config.test_batch_size = config.test_batch_size // config.gradient_accumulation_steps
        config.tasks[task.name]["train_batch_size"] =  config.tasks[task.name]["train_batch_size"] // config.gradient_accumulation_steps
        config.tasks[task.name]["test_batch_size"] = config.tasks[task.name]["test_batch_size"] // config.gradient_accumulation_steps
        # adjust to real training batch size
        utils.log("Training batch size: {}".format(config.tasks[task.name]["train_batch_size"]))

    calculated_num_train_optimization_steps = config.num_steps_in_one_epoch * config.epoch_number \
        if config.schedule_lr else -1
    if config.num_train_optimization_steps == 0:
      config.num_train_optimization_steps = calculated_num_train_optimization_steps
    else:
      utils.log("Overwriting the training steps to {} instead of {} because of the configuration".format(
        config.num_train_optimization_steps, calculated_num_train_optimization_steps))
    utils.log("Optimization steps : {}".format(config.num_train_optimization_steps))

    # Optimization
    no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
    param_optimizer = list(self.model.named_parameters())
    # for n, p in param_optimizer:
    #   print(n)
    optimizers = {}
    optim_type = ""

    if config.only_adam:
      """
      This optimization method can only support finetuning all.
      This is for the adaptation of [CEDR](https://github.com/Georgetown-IR-Lab/cedr).
      Will extend this for other settings, such as layer fixing
      """
      utils.log("Optimizing with Adam with {} and {} learninig rate".format(config.learning_rate, config.adam_learning_rate))
      if config.fix_embedding:
        modules_parameters = {"params": [p for n, p in param_optimizer if "bert" not in n and "word_embedding" not in n]}
      else:
        modules_parameters = {"params": [p for n, p in param_optimizer if "bert" not in n]}
      bert_optimizer_grouped_parameters = {'params': [p for n, p in param_optimizer if "bert" in n], 'lr': config.learning_rate}
      if len(bert_optimizer_grouped_parameters["params"]) == 0:
        utils.log("There is no BERT module in the model.")
        optimizers["optimizer"] = torch.optim.Adam([modules_parameters], lr=config.adam_learning_rate)
      else:
        optimizers["optimizer"] = torch.optim.Adam(
          [modules_parameters, bert_optimizer_grouped_parameters], lr=config.adam_learning_rate)
      optim_type = "only_adam"

    elif config.sep_optim:
      utils.log("Optimizing the module using Adam optimizer ..")
      modules_parameters = [p for n, p in param_optimizer if "bert" not in n]
      bert_optimizer_grouped_parameters = [
        {'params': [p for n, p in param_optimizer if ("bert" in n) and (not any(nd in n for nd in no_decay))],
         'weight_decay': 0.01},
        {'params': [p for n, p in param_optimizer if ("bert" in n) and (any(nd in n for nd in no_decay))],
         'weight_decay': 0.0}
      ]
      optimizers["module_optimizer"] = Adam(params=modules_parameters, lr=config.adam_learning_rate)
      optimizers["bert_optimizer"] = BertAdam(bert_optimizer_grouped_parameters,
                                              lr=config.learning_rate,
                                              warmup=config.warmup_proportion,
                                              schedule=config.schedule_method,
                                              t_total=config.num_train_optimization_steps)
      optim_type = "sep_optim"
    elif config.two_stage_optim:
      utils.log("Optimizing the module with two stage")
      modules_parameters = [p for n, p in param_optimizer if "bert" not in n]
      bert_optimizer_grouped_parameters = [
        {'params': [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)], 'weight_decay': 0.01},
        {'params': [p for n, p in param_optimizer if any(nd in n for nd in no_decay)], 'weight_decay': 0.0}]
      optimizers["module_optimizer"] = Adam(params=modules_parameters, lr=config.adam_learning_rate)
      optimizers["bert_optimizer"] = BertAdam(bert_optimizer_grouped_parameters,
                                              lr=config.learning_rate,
                                              warmup=config.warmup_proportion,
                                              schedule=config.schedule_method,
                                              t_total=config.num_train_optimization_steps)
      optim_type = "two_stage_optim"
    elif config.fix_bert:
      utils.log("Optimizing the module using Adam optimizer ..")
      modules_parameters = [p for n, p in param_optimizer if "bert" not in n]
      optimizers["module_optimizer"] = SGD(params=modules_parameters, lr=config.adam_learning_rate)
      optim_type = "fix_bert"
    elif config.fix_embedding:
      utils.log("Optimizing the model using one optimizer and fix embedding layer")
      bert_optimizer_grouped_parameters = [
        {'params': [p for n, p in param_optimizer if not any(nd in n for nd in no_decay) and ("word_embeddings" not in n)],
                'weight_decay': 0.01},
        {'params': [p for n, p in param_optimizer if any(nd in n for nd in no_decay) and ("word_embeddings" not in n)],
         'weight_decay': 0.0}]
      optimizers["bert_optimizer"] = BertAdam(bert_optimizer_grouped_parameters,
                                              lr=config.learning_rate,
                                              warmup=config.warmup_proportion,
                                              schedule=config.schedule_method,
                                              t_total=config.num_train_optimization_steps)
      optim_type = "normal"
    elif self.ext_config.encoder_configs.fix_embedding or self.ext_config.encoder_configs.fix_layers:
      utils.log("Fixing layers from config")
      print("Fix embedding {}".format(self.ext_config.encoder_configs.fix_embedding))
      print("Fix layers {}".format(self.ext_config.encoder_configs.fix_layers))
      self.ext_config: Configuration
      if config.encoder_type == 'xlmr':
        prefix = "layers"
        embed_prefix = "embed_tokens"
      elif config.encoder_type == "bert":
        prefix = "layer"
        embed_prefix = "word_embeddings"
      else:
        raise ValueError("Not supported encoder_type {}".format(config.encoder_type))
      bert_optimizer_grouped_parameters = [
        {'params': [],
         'weight_decay': 0.01},
        {'params': [],
        'weight_decay': 0.0}]

      for n, p in param_optimizer:
        if any(nd in n for nd in no_decay):
          # Checking embedding
          if embed_prefix in n:
            if not self.ext_config.encoder_configs.fix_embedding:
              bert_optimizer_grouped_parameters[1]['params'].append(p)
            else:
              print("Skip {}".format(n))
          if not any(".{}.{}.".format(prefix, l) in n for l in self.ext_config.encoder_configs.fix_layers):
            bert_optimizer_grouped_parameters[1]['params'].append(p)
          else:
            print("Skip {}".format(n))

        else:
          if embed_prefix in n:
            if not self.ext_config.encoder_configs.fix_embedding:
              bert_optimizer_grouped_parameters[0]['params'].append(p)
            else:
              print("Skip {}".format(n))
          if not any(".{}.{}.".format(prefix, l) in n for l in self.ext_config.encoder_configs.fix_layers):
            bert_optimizer_grouped_parameters[0]['params'].append(p)
          else:
            print("Skip {}".format(n))
      optimizers["bert_optimizer"] = BertAdam(bert_optimizer_grouped_parameters,
                                              lr=config.learning_rate,
                                              warmup=config.warmup_proportion,
                                              schedule=config.schedule_method,
                                              t_total=config.num_train_optimization_steps)
      optim_type = "normal"

    else:
      utils.log("Optimizing the model using one optimizer")
      bert_optimizer_grouped_parameters = [
        {'params': [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)], 'weight_decay': 0.01},
        {'params': [p for n, p in param_optimizer if any(nd in n for nd in no_decay)], 'weight_decay': 0.0}]
      optimizers["bert_optimizer"] = BertAdam(bert_optimizer_grouped_parameters,
                                lr=config.learning_rate,
                                warmup=config.warmup_proportion,
                                schedule=config.schedule_method,
                                t_total=config.num_train_optimization_steps)
      optim_type = "normal"

    self.optimizer = MultipleOptimizer(optim=optimizers, optim_type=optim_type)

    self.global_step_labeled = 0
    self.global_step_unlabeled = 0


  def flip_the_coin(self, step):
    if np.random.rand() < min(0.6, 1 / 10000 * step):
      return True
    return False

  def gen_masks(self, label_ids):
    sample = torch.rand(label_ids.size())
    return sample

  def train_labeled_abstract(self, mb: MiniBatch, step):
    self.model.train()

    inputs = mb.generate_input(device=self.device, use_label=True)
    if "input_ids" in inputs and inputs["input_ids"] is not None and inputs["input_ids"].size(0) == 0:
      utils.log("Zero Batch")
      return 0
    if "_input_token_ids" in inputs and inputs["_input_token_ids"] is not None and inputs["_input_token_ids"].size(0) == 0:
      utils.log("Zero Batch")
      return 0

    outputs = self.model(**inputs)

    # TODO: Slow process Migrating Interface ...
    if isinstance(outputs, dict):
      loss = outputs[mb.task_name]["loss"]
    else:
      if self.config.output_attentions:
        loss, _, _ = outputs
      else:
        loss, _ = outputs

    loss = mb.loss_weight * loss

    if self.config.gradient_accumulation_steps > 1:
      loss = loss / self.config.gradient_accumulation_steps
    loss.backward()
    if (step + 1) % self.config.gradient_accumulation_steps == 0:
      # TODO: a quick fix
      if not hasattr(mb, "task_name") or mb.task_name not in ["squad11", "squad20"]:
        nn.utils.clip_grad_norm_(self.model.parameters(), self.config.grad_clip)
      self.optimizer.step()
      self.optimizer.zero_grad()
      self.global_step_labeled += 1
    return loss.item()

  def train_discriminator(self, labeled_mb, unlabeled_mb):
    self.model.train()
    self.adversarial_agent.train()
    labeled_inputs = labeled_mb.generate_input(device=self.device, use_label=False)
    unlabeled_inputs = unlabeled_mb.generate_input(device=self.device, use_label=False)
    # We want to overwrite the task name here.
    # We only want to get the feature of encoder,
    # we do not want to do the decoder.
    labeled_inputs["task_name"] = unlabeled_inputs["task_name"]
    # unlabeled_inputs["task_name"] = "encoding"
    labeled_features = self.model(**labeled_inputs)
    unlabeled_features = self.model(**unlabeled_inputs)
    # we delay the detach operation in the discriminator
    labeled_loss, unlabeled_loss, real_acc, fake_acc = self.adversarial_agent.update(
        labeled_features[labeled_inputs["task_name"]]["logits"].detach(),
        unlabeled_features[unlabeled_inputs["task_name"]]["logits"].detach(),
        1.0, 0.0)
    # We name it as `logits` because we aggregate sequence of vector into a vector.
    # TODO: change it if it's not binary case
    # In the update operation, the following things will be done
    # 1. Discriminator will be called
    # 2. Loss will be calculated
    # 3. Backward function will be called
    # 4. Optimizer step function will be called
    return labeled_loss, unlabeled_loss, real_acc, fake_acc


  def train_generator(self, labeled_mb: MiniBatch, unlabeled_mb: MiniBatch):
    self.model.train()
    self.adversarial_agent.train()
    labeled_inputs = labeled_mb.generate_input(device=self.device, use_label=True)
    unlabeled_inputs = unlabeled_mb.generate_input(device=self.device, use_label=False)

    # A quick hack on reuse encoder output, will reorganize this later
    original_labeled_input_task_name = labeled_inputs["task_name"]
    labeled_inputs["task_name"] = ",".join([labeled_inputs["task_name"], unlabeled_inputs["task_name"]])
    labeled_outputs = self.model(**labeled_inputs)
    labeled_loss = labeled_outputs[original_labeled_input_task_name]["loss"]
    # unlabeled_inputs["task_name"] = "encoding"
    unlabeled_outputs = self.model(**unlabeled_inputs)
    discriminator_loss = self.adversarial_agent.gen_loss(
      labeled_outputs[unlabeled_inputs["task_name"]]["logits"],
      unlabeled_outputs[unlabeled_inputs["task_name"]]["logits"],
      1.0, 0.0)
    loss = labeled_loss + discriminator_loss
    loss.backward()
    nn.utils.clip_grad_norm_(self.model.parameters(), self.config.grad_clip)
    # For the adversarial training, we now do not support the loss accumulation
    self.optimizer.step()
    self.optimizer.zero_grad()
    return labeled_loss.item(), discriminator_loss.item()

  def test_abstract(self, mb):
    self.model.eval()

    inputs = mb.generate_input(self.device, use_label=False)

    with torch.no_grad():
      results = self.model(**inputs)
    if isinstance(results, dict):
      return results
    else:
      if self.config.output_attentions:
        results, attention_map = results
        # list(batch_size, num_heads, sent_length, sent_length) = layer
        attention_map = torch.stack(attention_map, dim=0).transpose(0, 1)
      # (layer, batch, head, length, length)
      if self.config.output_attentions:
        return results, attention_map.cpu().numpy()
      else:
        return results

  def analyze(self, mb, head_mask, params):
    self.model.eval()

    inputs = mb.generate_input(self.device, use_label=False)

    inputs["extra_args"]["head_mask"] = head_mask
    outputs = self.model(**inputs)
    # We assume here is loss, logits, and attention_map
    loss, logits, all_attentions = outputs
    loss.backward()  # Backpropagate to populate the gradients in the head mask
    # Compute Entropy

    for layer, attn in enumerate(all_attentions):
      masked_entropy = entropy(attn.detach()) * inputs["input_mask"].float().unsqueeze(1)
      params["attn_entropy"][layer] += masked_entropy.sum(-1).sum(0).detach()

    params["head_importance"] += head_mask.grad.abs().detach()
    params["total_token"] += inputs["input_mask"].float().detach().sum().data

    return logits, torch.stack(all_attentions, dim=0).transpose(0, 1).detach().data.cpu().numpy()


  def run_teacher_abstract(self, mb: MiniBatch):
    self.teacher.eval()
    inputs = mb.generate_input(device=self.device, use_label=False)
    with torch.no_grad():
      results = self.teacher(**inputs)
      mb.teacher_predictions = results


  def train_unlabeled_abstract(self, mb: MiniBatch, step):
    self.model.train()
    inputs = mb.generate_input(device=self.device, use_label=False)
    outputs = self.model(**inputs,
            teacher_predictions=mb.teacher_predictions)

    loss, _ = outputs

    loss = mb.loss_weight * loss

    if self.config.gradient_accumulation_steps > 1:
      loss = loss / self.config.gradient_accumulation_steps
    loss.backward()
    if (step + 1) % self.config.gradient_accumulation_steps == 0:
      # TODO: a quick fix
      if not hasattr(mb, "task_name") or mb.task_name not in ["squad11", "squad20"]:
        nn.utils.clip_grad_norm_(self.model.parameters(), self.config.grad_clip)
      self.optimizer.step()
      self.optimizer.zero_grad()
      self.global_step_unlabeled += 1
    return loss.item()



from relogic.logickit.dataflow import TASK_TO_DATAFLOW_CLASS_MAP, ECPDataFlow
from relogic.logickit.base.constants import ECP_TASK
from types import SimpleNamespace
import json
from relogic.logickit.tokenizer.tokenization import BertTokenizer
import relogic.utils.crash_on_ipy

config = SimpleNamespace(**{
  "buckets": [(0, 90), (90, 150), (150, 200), (200, 450)],
  "max_seq_length": 450,
  "label_mapping_path": "data/preprocessed_data/binary_classification.json"})

tokenizers = {
  "BPE": BertTokenizer.from_pretrained("bert-base-multilingual-cased", lang="zh")
}

dataflow: ECPDataFlow = TASK_TO_DATAFLOW_CLASS_MAP[ECP_TASK](
  task_name=ECP_TASK,
  config=config,
  tokenizers=tokenizers,
  label_mapping=json.load(open(config.label_mapping_path)))


examples = [{
  "tokens": ['当', '我', '看', '到', '建', '议', '被', '采', '纳', ',', '部', '委', '领', '导', '写', '给', '我', '的', '回', '信', '时', ',', '我', '知', '道', '我', '正', '在', '为', '这', '个', '国', '家', '的', '发', '展', '尽', '着', '一', '份', '力', '量', ',', '2', '7', '日', ',', '河', '北', '省', '邢', '台', '钢', '铁', '有', '限', '公', '司', '的', '普', '通', '工', '人', '白', '金', '跃', ',', '拿', '着', '历', '年', '来', '国', '家', '各', '部', '委', '反', '馈', '给', '他', '的', '感', '谢', '信', ',', '激', '动', '地', '对', '中', '新', '网', '记', '者', '说', ',', '2', '7', '年', '来', ',', '国', '家', '公', '安', '部', '国', '家', '工', '商', '总', '局', '国', '家', '科', '学', '技', '术', '委', '员', '会', '科', '技', '部', '卫', '生', '部', '国', '家', '发', '展', '改', '革', '委', '员', '会', '等', '部', '委', '均', '接', '受', '并', '采', '纳', '过', '的', '我', '的', '建', '议'],
  "labels": [(86, 96, 102, 152)],
  "clause_spans": [(0, 9), (10, 21), (22, 42), (43, 46), (47, 66), (67, 85), (86, 96), (97, 101), (102, 152)]
}]

# Some problem with Chinese Indexing
# for token list (Chinese characters), they will be combined into a string without space.
# Until now the spans index is still correct.
# But after tokenization, something goes wrong. Because some numbers will be merged (several digits will be merged)
# Then the index information is wrong here.


dataflow.update_with_jsons(examples)

for mb in dataflow.get_minibatches(minibatch_size=1):
  print(mb)

raise NotImplementedError("You can start to play with data")





from relogic.logickit.inference.modeling import BertPreTrainedModel, BertModel
from relogic.logickit.inference.modeling_xlm import XLMPreTrainedModel, XLMModel
from relogic.logickit.modules.contextualizers.highway_lstm import HighwayLSTM
from relogic.utils.file_utils import cached_path
import torch
import torch.nn as nn
import numpy as np
import logging

logger = logging.getLogger(__name__)

def get_encoder(encoder_type):
  if encoder_type == "bert":
    return Encoder
  if encoder_type == "xlm":
    return XLMEncoder
  if encoder_type == "xlmr":
    return XLMRobertaEncoder
  if encoder_type == "lstm":
    return LSTMEncoder
  if encoder_type == "embedding":
    return PretrainedWordEmbedding

PRETRAINED_VECTOR_ARCHIVE_MAP = {
  "fasttext-en": "https://git.uwaterloo.ca/p8shi/data-server/raw/master/embeddings/wiki-news-300d-1M.npy"
}

class PretrainedWordEmbedding(nn.Module):
  def __init__(self, pretrained_model_name_or_path, embedding_file_path):
    super().__init__()
    embedding_data = torch.from_numpy(np.load(embedding_file_path))
    self.word_embedding = nn.Embedding(embedding_data.size(0), embedding_data.size(1))
    self.word_embedding.weight.data.copy_(embedding_data)
    self.word_embedding.weight.requires_grad = False
    del embedding_data
    print("Loadding embedding from {}".format(embedding_file_path))

  def forward(self, *inputs, **kwargs):
    input_token_ids = kwargs.pop("_input_token_ids")
    input_embedding = self.word_embedding(input_token_ids)
    return input_embedding

  @classmethod
  def from_pretrained(cls, pretrained_model_name_or_path, cache_dir=None, output_attentions=False):
    if pretrained_model_name_or_path in PRETRAINED_VECTOR_ARCHIVE_MAP:
      embedding_file = PRETRAINED_VECTOR_ARCHIVE_MAP[pretrained_model_name_or_path]
    else:
      embedding_file = pretrained_model_name_or_path
    try:
      resolved_embedding_file = cached_path(embedding_file, cache_dir=cache_dir)
    except EnvironmentError:
      logger.error(
        "Model name '{}' was not found in model name list ({}). "
        "We assumed '{}' was a path or url but couldn't find any file "
        "associated to this path or url.".format(
          pretrained_model_name_or_path,
          ', '.join(PRETRAINED_VECTOR_ARCHIVE_MAP.keys()),
          embedding_file))
      return None
    if resolved_embedding_file == embedding_file:
      logger.info("will load embedding file from {}".format(embedding_file))
    else:
      logger.info("will load embedding file {} from cache at {}".format(
        embedding_file, resolved_embedding_file))
    return cls(pretrained_model_name_or_path, embedding_file_path=resolved_embedding_file)


class CNNEncoder(nn.Module):
  def __init__(self, pretrained_model_name_or_path, embedding_file_path):
    super().__init__()
    self.encoder = None
    embedding_data = torch.from_numpy(np.load(embedding_file_path))
    self.word_embedding = nn.Embedding(embedding_data.size(0), embedding_data.size(1))
    self.word_embedding.weight.data.copy_(embedding_data)
    del embedding_data
    print("Loading embedding from {}".format(embedding_file_path))

  def forward(self, *input, **kwargs):
    pass

  @classmethod
  def from_pretrained(cls, pretrained_model_name_or_path, cache_dir=None, output_attentions=False):
    if pretrained_model_name_or_path in PRETRAINED_VECTOR_ARCHIVE_MAP:
      embedding_file = PRETRAINED_VECTOR_ARCHIVE_MAP[pretrained_model_name_or_path]
    else:
      embedding_file = pretrained_model_name_or_path
    try:
      resolved_embedding_file = cached_path(embedding_file, cache_dir=cache_dir)
    except EnvironmentError:
      logger.error(
        "Model name '{}' was not found in model name list ({}). "
        "We assumed '{}' was a path or url but couldn't find any file "
        "associated to this path or url.".format(
          pretrained_model_name_or_path,
          ', '.join(PRETRAINED_VECTOR_ARCHIVE_MAP.keys()),
          embedding_file))
      return None
    if resolved_embedding_file == embedding_file:
      logger.info("will load embedding file from {}".format(embedding_file))
    else:
      logger.info("will load embedding file {} from cache at {}".format(
        embedding_file, resolved_embedding_file))
    return cls(pretrained_model_name_or_path, embedding_file_path=resolved_embedding_file)


class LSTMEncoder(nn.Module):
  def __init__(self, pretrained_model_name_or_path, embedding_file_path):
    super().__init__()
    self.encoder = HighwayLSTM(num_layers=2, input_size=300, hidden_size=200, layer_dropout=0.2)
    embedding_data = torch.from_numpy(np.load(embedding_file_path))
    self.word_embedding = nn.Embedding(embedding_data.size(0), embedding_data.size(1))
    self.word_embedding.weight.data.copy_(embedding_data)
    del embedding_data
    print("Loading embedding from {}".format(embedding_file_path))

  def forward(self, *inputs, **kwargs):
    input_token_ids = kwargs.pop("_input_token_ids")
    token_length = kwargs.pop("_token_length")
    input_embedding = self.word_embedding(input_token_ids)
    sequence_output = self.encoder(inputs=input_embedding, lengths=token_length)
    return sequence_output


  @classmethod
  def from_pretrained(cls, pretrained_model_name_or_path, cache_dir=None, output_attentions=False):
    if pretrained_model_name_or_path in PRETRAINED_VECTOR_ARCHIVE_MAP:
      embedding_file = PRETRAINED_VECTOR_ARCHIVE_MAP[pretrained_model_name_or_path]
    else:
      embedding_file = pretrained_model_name_or_path
    try:
      resolved_embedding_file = cached_path(embedding_file, cache_dir=cache_dir)
    except EnvironmentError:
      logger.error(
        "Model name '{}' was not found in model name list ({}). "
        "We assumed '{}' was a path or url but couldn't find any file "
        "associated to this path or url.".format(
          pretrained_model_name_or_path,
          ', '.join(PRETRAINED_VECTOR_ARCHIVE_MAP.keys()),
          embedding_file))
      return None
    if resolved_embedding_file == embedding_file:
      logger.info("will load embedding file from {}".format(embedding_file))
    else:
      logger.info("will load embedding file {} from cache at {}".format(
        embedding_file, resolved_embedding_file))
    return cls(pretrained_model_name_or_path, embedding_file_path=resolved_embedding_file)

class XLMRobertaEncoder(nn.Module):
  def __init__(self, pretrained_model_name_or_path):
    super().__init__()
    self.xlmr = torch.hub.load('pytorch/fairseq', pretrained_model_name_or_path)

  def forward(self, input_ids, **kwargs):
    sequence_output = self.xlmr.extract_features(input_ids)
    return sequence_output

  @classmethod
  def from_pretrained(cls, pretrained_model_name_or_path, output_attentions=False):
    assert(not output_attentions)
    # Need to figure out how to output attention from XLMR
    return cls(pretrained_model_name_or_path)


class XLMEncoder(nn.Module):
  def __init__(self, pretrained_model_name_or_path):
    super(XLMEncoder, self).__init__()
    self.xlm = XLMModel.from_pretrained(
      pretrained_model_name_or_path)

  def forward(self,
              input_ids,
              attention_mask=None,
              langs=None,
              token_type_ids=None,
              position_ids=None,
              lengths=None,
              cache=None,
              head_mask=None,
              output_all_encoder_layers=False,
              selected_non_final_layers=None,
              route_path=None,
              no_dropout=True,
              **kwargs):
    """

    :param input_ids:
    :param attention_mask:
    :param langs: (batch_size, sentence_length)
    :param token_type_ids:
    :param position_ids:
    :param lengths:
    :param cache:
    :param head_mask:
    :param output_all_encoder_layers:
    :param selected_non_final_layers:
    :param route_path:
    :param no_dropout:
    :return:
    """
    sequence_output = self.xlm(input_ids=input_ids,
             attention_mask=attention_mask,
             langs=langs,
             token_type_ids=token_type_ids,
             position_ids=position_ids,
             lengths=lengths,
             cache=cache,
             head_mask=head_mask)
    return sequence_output

  @classmethod
  def from_pretrained(cls, pretrained_model_name_or_path, output_attentions=False):
    assert(not output_attentions)
    # TODO: Figure out how to output attention map from XLM
    return cls(pretrained_model_name_or_path)




class Encoder(BertPreTrainedModel):
  def __init__(self, config, output_attentions=False):
    super(Encoder, self).__init__(config)
    self._config = config
    self.output_attentions = output_attentions
    self.bert = BertModel(config, output_attentions=output_attentions)
    self.dropout = nn.Dropout(config.hidden_dropout_prob)
    self.apply(self.init_bert_weights)

  def forward(self,
              input_ids,
              token_type_ids=None,
              attention_mask=None,
              head_mask=None,
              output_all_encoded_layers=False,
              token_level_attention_mask=None,
              selected_non_final_layers=None,
              route_path=None,
              no_dropout=False,
              **kwargs):
    sequence_output = self.bert(
      input_ids=input_ids,
      token_type_ids=token_type_ids,
      attention_mask=attention_mask,
      head_mask=head_mask,
      token_level_attention_mask=token_level_attention_mask,
      output_all_encoded_layers=output_all_encoded_layers,
      selected_non_final_layers=selected_non_final_layers)
    if self.output_attentions:
      attention_map, sequence_output, _ = sequence_output
    else:
      sequence_output, _ = sequence_output
    if not no_dropout:
      if output_all_encoded_layers or selected_non_final_layers is not None:
        sequence_output = [self.dropout(seq) for seq in sequence_output]
      else:
        sequence_output = self.dropout(sequence_output)
    if self.output_attentions:
      return sequence_output, attention_map
    else:
      return sequence_output


if __name__ == "__main__":
  import torch
  # model = BertModel.from_pretrained("bert-large-uncased-whole-word-masking")
  model = LSTMEncoder.from_pretrained("fasttext-en")
  # layer1 = model.encoder.layer[0]
  # print(layer1[0].attention.self.value.weight.data - layer1[1].attention.self.value.weight.data)
  feature = model(_input_token_ids=torch.zeros(2, 5, dtype=torch.long),
                 _token_length=torch.LongTensor([3, 5]))

  print(feature)

  # encoder = XLMEncoder("xlm-mlm-tlm-xnli15-1024")
  # feature = encoder(input_ids=torch.zeros(1, 5, dtype=torch.long))
  # print(feature)



import torch.nn as nn
import torch


class LR(nn.Module):
  def __init__(self, config):
    super().__init__()
    self.model = nn.Linear(in_features=config.hidden_size, out_features=1)

  def forward(self, *inputs, **kwargs):
    feature = kwargs.pop("features")
    if feature.dim() == 3:
      feature = torch.mean(feature, dim=1)
    output = self.model(feature)
    return output




from dataclasses import dataclass, field
from typing import List, Union

from relogic.structures.structure import Structure

@dataclass
class Pointwise(Structure):
  text_a: str = None
  text_b: str = None


import json
from relogic.logickit.tokenizer.tokenization import whitespace_tokenize
import argparse

class SquadExample(object):
  """
  A single training/test example for the Squad dataset.
  For examples without an answer, the start and end position are -1.
  """

  def __init__(self,
               qas_id,
               question_text,
               doc_tokens,
               orig_answer_text=None,
               start_position=None,
               end_position=None,
               is_impossible=None):
    self.qas_id = qas_id
    self.question_text = question_text
    self.doc_tokens = doc_tokens
    self.orig_answer_text = orig_answer_text
    self.start_position = start_position
    self.end_position = end_position
    self.is_impossible = is_impossible

  def __str__(self):
    return self.__dict__

def read_squad_examples(input_file, is_training, version_2_with_negative):
  """Read a SQuAD json file into a list of SquadExample."""
  with open(input_file, "r", encoding='utf-8') as reader:
    input_data = json.load(reader)["data"]

  def is_whitespace(c):
    if c == " " or c == "\t" or c == "\r" or c == "\n" or ord(c) == 0x202F:
      return True
    return False

  examples = []
  for entry in input_data:
    for paragraph in entry["paragraphs"]:
      paragraph_text = paragraph["context"]
      doc_tokens = []
      char_to_word_offset = []
      prev_is_whitespace = True
      for c in paragraph_text:
        if is_whitespace(c):
          prev_is_whitespace = True
        else:
          if prev_is_whitespace:
            doc_tokens.append(c)
          else:
            doc_tokens[-1] += c
          prev_is_whitespace = False
        char_to_word_offset.append(len(doc_tokens) - 1)

      for qa in paragraph["qas"]:
        qas_id = qa["id"]
        question_text = qa["question"]
        start_position = None
        end_position = None
        orig_answer_text = None
        is_impossible = False
        if is_training:
          if version_2_with_negative:
            is_impossible = qa["is_impossible"]
          if (len(qa["answers"]) != 1) and (not is_impossible):
            raise ValueError(
              "For training, each question should have exactly 1 answer.")
          if not is_impossible:
            answer = qa["answers"][0]
            orig_answer_text = answer["text"]
            answer_offset = answer["answer_start"]
            answer_length = len(orig_answer_text)
            start_position = char_to_word_offset[answer_offset]
            end_position = char_to_word_offset[answer_offset + answer_length - 1]
            # Only add answers where the text can be exactly recovered from the
            # document. If this CAN'T happen it's likely due to weird Unicode
            # stuff so we will just skip the example.
            #
            # Note that this means for training mode, every example is NOT
            # guaranteed to be preserved.
            actual_text = " ".join(doc_tokens[start_position:(end_position + 1)])
            cleaned_answer_text = " ".join(
              whitespace_tokenize(orig_answer_text))
            if actual_text.find(cleaned_answer_text) == -1:
              print("Could not find answer: '%s' vs. '%s'",
                             actual_text, cleaned_answer_text)
              continue
          else:
            start_position = -1
            end_position = -1
            orig_answer_text = ""

        example = SquadExample(
          qas_id=qas_id,
          question_text=question_text,
          doc_tokens=doc_tokens,
          orig_answer_text=orig_answer_text,
          start_position=start_position,
          end_position=end_position,
          is_impossible=is_impossible)
        examples.append(example)
  return examples

if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("--input_file")
  parser.add_argument("--output_file")
  parser.add_argument("--is_training", default=False, action="store_true")
  parser.add_argument("--with_negative", default=False, action="store_true")
  args = parser.parse_args()
  with open(args.output_file, "w") as fout:
    examples = read_squad_examples(args.input_file, args.is_training, args.with_negative)
    for example in examples:
      fout.write(json.dumps({
        "guid": example.qas_id,
        "question_text": example.question_text,
        "doc_tokens": example.doc_tokens,
        "start_position": example.start_position,
        "end_position": example.end_position,
        "is_impossible": example.is_impossible,
        "orig_answer_text": example.orig_answer_text
      }) + "\n")



from relogic.graphkit.linking.simple_entity_linker import SimpleEntityLinker
from relogic.structures.span import Span

linker = SimpleEntityLinker({
  "en": "/data/lctan/Nesoi/indexes/lucene-index.en-mention-table.pos+docvectors+rawdocs/"})
span = Span(text="florida")
linker.link([span])

print(span.ranked_uris[0])



import abc
from relogic.logickit.dataflow import DataFlow


class Task(object, metaclass=abc.ABCMeta):
  def __init__(self, config, name, loader):
    self.config = config
    self.name = name
    self.loader = loader
    self.has_module = True
    self.has_scorer = True
    if config.mode == 'train' or config.mode == 'finetune':
      self.train_set = self.loader.get_dataset("train")
    else:
      self.train_set = None
    if config.mode != "deployment":
      self.val_set = self.loader.get_dataset("dev" if (
            config.mode == 'train' or config.mode == 'valid' or config.mode == "finetune") else "test")
    else:
      self.val_set = None
    try:
      self.dataset: DataFlow = self.loader.get_dataflow()
    except:
      self.dataset = None

  @abc.abstractmethod
  def get_module(self):
    pass

  @abc.abstractmethod
  def get_scorer(self, dump_to_file=None):
    pass






import argparse
import json

def preprocess(input_file, output_file):
  examples = []
  tokens, labels, arcs, pos = [], [], [], []
  with open(input_file) as fin:
    for line in fin:
      if line == "\n":
        examples.append({
          "tokens": tokens,
          "labels": labels,
          "arcs": arcs,
          "pos": pos
        })
        tokens, labels, arcs, pos = [], [], [], []
      else:
        items = line.strip().split('\t')
        tokens.append(items[1])
        labels.append(items[7].split(":")[0])
        pos.append(items[4])
        arcs.append(int(items[6]))
  with open(output_file, 'w') as fout:
    for example in examples:
      fout.write(json.dumps(example)+"\n")

if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("--input_file", type=str)
  parser.add_argument("--output_file", type=str)
  args = parser.parse_args()
  preprocess(args.input_file, args.output_file)


import torch
import torch.nn as nn

class MatchingModule(nn.Module):
  def __init__(self, config, task_name, n_classes):
    super(MatchingModule, self).__init__()
    self.config = config
    self.task_name = task_name
    if self.config.regression:
      self.n_classes = 1
    else:
      self.n_classes = n_classes
    if self.config.ir_siamese:
      hidden_size = config.hidden_size * 2
    else:
      hidden_size = config.hidden_size
    self.to_logits = nn.Linear(hidden_size, self.n_classes)

  # def forward(self, input, input_mask=None, segment_ids=None, extra_args=None, **kwargs):
  def forward(self, *inputs, **kwargs):
    if self.config.ir_siamese:
      a_features = kwargs.pop("a_features")
      b_features = kwargs.pop("b_features")
      logits = self.to_logits(torch.cat([a_features[:, 0], b_features[:, 0]], dim=-1))
    else:
      features = kwargs.pop("features")
      logits = self.to_logits(features[:,0])
    return logits


import torch.nn as nn
import torch
from relogic.logickit.modules.span_extractors.endpoint_span_extractor import EndpointSpanExtractor
from relogic.logickit.modules.span_extractors.self_attentive_span_extractor import SelfAttentiveSpanExtractor
from relogic.logickit.modules.span_extractors.average_span_extractor import AverageSpanExtractor
from relogic.logickit.modules.span_extractors.attentive_span_extractor import AttentiveSpanExtractor
from relogic.logickit.modules.pruner import Pruner
from relogic.logickit.inference.modeling import gelu
from relogic.logickit.utils import utils
from relogic.logickit.base.constants import SPAN_REPR_KENTON_LEE, SPAN_REPR_AVE_MAX, SPAN_REPR_AVE

try:
  from torch_geometric.nn import GCNConv

  class GraphNet(nn.Module):
    def __init__(self, feature_size, hidden_size):
      super(GraphNet, self).__init__()
      self.conv1 = GCNConv(feature_size, hidden_size)
      self.conv2 = GCNConv(hidden_size, hidden_size)

    def forward(self, x, edge_index):
      x = self.conv1(x, edge_index)
      x = torch.relu(x)
      x = torch.dropout(x, training=self.training)
      x = self.conv2(x, edge_index)
      return x
except:
  pass

class LayerNorm(nn.Module):
  def __init__(self, hidden_size, eps=1e-12):
    """Construct a layernorm module in the TF style (epsilon inside the square root).
    """
    super(LayerNorm, self).__init__()
    self.weight = nn.Parameter(torch.ones(hidden_size))
    self.bias = nn.Parameter(torch.zeros(hidden_size))
    self.variance_epsilon = eps

  def forward(self, x):
    u = x.mean(-1, keepdim=True)
    s = (x - u).pow(2).mean(-1, keepdim=True)
    x = (x - u) / torch.sqrt(s + self.variance_epsilon)
    return self.weight * x + self.bias

class JointSRLModule(nn.Module):

  def __init__(self, config, task_name, n_classes):
    super(JointSRLModule, self).__init__()
    self.config = config
    self.task_name = task_name
    self.n_classes = n_classes
    self.endpoint_span_extractor = EndpointSpanExtractor(
      input_dim=config.hidden_size,
      num_width_embeddings=config.num_width_embeddings,
      span_width_embedding_dim=config.span_width_embedding_dim)
    self.endpoint_predicate_span_extractor = EndpointSpanExtractor(
      input_dim=config.hidden_size)

    hidden_size = config.hidden_size

    self.attentive_span_extractor = AttentiveSpanExtractor(input_dim=hidden_size)
    # SelfAttentiveSpanExtractor(input_dim=config.hidden_size)
    # self.average_span_extractor = AverageSpanExtractor(input_dim=config.hidden_size)

    if self.config.srl_arg_span_repr == SPAN_REPR_AVE:
      arg_span_dim = hidden_size
    elif self.config.srl_arg_span_repr == SPAN_REPR_AVE_MAX:
      arg_span_dim = hidden_size * 2
    elif self.config.srl_arg_span_repr == SPAN_REPR_KENTON_LEE:
      arg_span_dim = hidden_size * 3 + config.span_width_embedding_dim
    else:
      raise ValueError("The span repr is not defined {}".format(self.config.srl_arg_span_repr))

    if self.config.srl_pred_span_repr == SPAN_REPR_AVE:
      pred_span_dim = hidden_size
    elif self.config.srl_pred_span_repr == SPAN_REPR_AVE_MAX:
      pred_span_dim = hidden_size * 2
    elif self.config.srl_pred_span_repr == SPAN_REPR_KENTON_LEE:
      pred_span_dim = hidden_size * 2
    else:
      raise ValueError("The span repr is not defined {}".format(self.config.srl_pred_span_repr))


    arg_feedforward_scorer = nn.Sequential(
      nn.Linear(in_features=arg_span_dim,
                out_features=config.hidden_size),
      nn.ReLU(),
      # LayerNorm(hidden_size=config.hidden_size),
      nn.Linear(in_features=config.hidden_size, out_features=1))
    self.mention_pruner = Pruner(arg_feedforward_scorer)
    predicate_feedforward_scorer = nn.Sequential(
      nn.Linear(in_features=pred_span_dim, out_features=config.hidden_size),
      nn.ReLU(),
      # LayerNorm(hidden_size=config.hidden_size),
      nn.Linear(in_features=config.hidden_size, out_features=1))
    self.predicate_pruner = Pruner(predicate_feedforward_scorer)

    self.semantic_role_predictor = nn.Sequential(
      nn.Linear(in_features=pred_span_dim + arg_span_dim + (300 if hasattr(config, "srl_use_label_embedding") and config.srl_use_label_embedding else 0),
                out_features=hidden_size),
      nn.ReLU(),
      # LayerNorm(hidden_size=config.hidden_size),
      nn.Linear(in_features=hidden_size, out_features=self.n_classes-1))
    self.constant_pred = nn.Parameter(torch.LongTensor([15]), requires_grad=False)
    self.constant_arg = nn.Parameter(torch.LongTensor([30]), requires_grad=False)
    # hard code here for now
    # TODO: make it to take the size from the vocab file
    if hasattr(config, "srl_use_label_embedding") and config.srl_use_label_embedding:
      self.label_embdding_key = nn.Embedding(27038, 300)
      self.label_embdding_value = nn.Embedding(27038, 300)
      self.pair_to_embedding = nn.Sequential(
        nn.Linear(in_features=arg_span_dim + pred_span_dim, out_features=config.hidden_size),
        nn.ReLU(),
        nn.Linear(in_features=config.hidden_size, out_features=300))
      self.to_attention = nn.Sequential(
        nn.Linear(in_features=600, out_features=100),
        nn.ReLU(),
        nn.Linear(in_features=100, out_features=1))
      # self.to_key = nn.Linear(in_features=300, out_features=300)
      # self.to_value = nn.Linear(in_features=300, out_features=300)
    # Hard code here for postag size
    if hasattr(config, "srl_compute_pos_tag_loss") and config.srl_compute_pos_tag_loss:
      self.to_pos_tag = nn.Sequential(
        nn.Linear(in_features=config.hidden_size, out_features=300),
        nn.ReLU(),
        nn.Linear(in_features=300, out_features=46)
      )
      # self.to_pos_tag = nn.Linear(in_features=config.hidden_size, out_features=46)

  def forward(self, *inputs, **kwargs):
    features = kwargs.pop("features")
    pos_tag_features = features[0]
    pred_features = features[1]
    arg_features = features[2]
    features = features[3]
    # TODO : quick fix for lstm testing
    arg_candidates = kwargs.pop("arg_candidates")
    predicate_candidates = kwargs.pop("predicate_candidates")

    num_arg_spans = arg_candidates.size(1)
    num_preds_spans = predicate_candidates.size(1)

    # Span padding is (1, 0). So we use [:, :, 1] > 0 can determine
    # the the span is padding or not.
    arg_mask = (arg_candidates[:, :, 1] > 0).float()
    predicate_mask = (predicate_candidates[:, :, 1] > 0).float()

    if self.config.srl_arg_span_repr == SPAN_REPR_KENTON_LEE:
      endpoint_span_embeddings = self.endpoint_span_extractor(
        sequence_tensor=features, span_indices=arg_candidates, span_indices_mask=arg_mask)

      attended_span_embeddings = self.attentive_span_extractor(
        sequence_tensor=features, value_tensor=arg_features, span_indices=arg_candidates, span_indices_mask=arg_mask)

      arg_embeddings = torch.cat([endpoint_span_embeddings, attended_span_embeddings], -1)
    elif self.config.srl_arg_span_repr == SPAN_REPR_AVE_MAX:
      average_span_embeddings = self.average_span_extractor(
        sequence_tensor=features, span_indices=arg_candidates, span_indices_mask=arg_mask
      )
      max_pooled_span_embedding = self.max_pooling_extractor(

      )
      arg_embeddings = torch.cat([average_span_embeddings, max_pooled_span_embedding], -1)
    elif self.config.srl_arg_span_repr == SPAN_REPR_AVE:
      arg_embeddings = self.average_span_extractor(
        sequence_tensor=features, span_indices=arg_candidates, span_indices_mask=arg_mask
      )
    else:
      raise ValueError("The span repr is not defined {}".format(self.config.srl_arg_span_repr))
    if self.config.srl_use_gold_argument:
      num_spans_to_keep = torch.sum(arg_mask, dim=-1).long()
    else:
      num_spans_to_keep = (torch.sum(arg_mask, dim=-1) * 0.8).long()
      num_spans_to_keep = torch.min(num_spans_to_keep, self.constant_arg) # quick fix

    (top_arg_span_embeddings, top_arg_span_mask,
     top_arg_span_indices, top_arg_span_mention_scores,
     arg_span_mention_full_scores) = self.mention_pruner(
      arg_embeddings, arg_mask, num_spans_to_keep)
    # top_arg_span_mask = (batch_size, max_arg_to_keep)

    if self.config.srl_pred_span_repr == SPAN_REPR_KENTON_LEE:
      pred_embeddings = self.endpoint_predicate_span_extractor(
        sequence_tensor=pred_features, span_indices=predicate_candidates, span_indices_mask=predicate_mask)
    elif self.config.srl_pred_span_repr == SPAN_REPR_AVE_MAX:
      average_pred_embeddings = self.average_sapn_extractor(
        sequence_tensor=features, span_indices=predicate_candidates, span_indices_mask=predicate_mask)
      max_pooled_pred_embeddings = self.max_pooling_extractor(

      )
      pred_embeddings = torch.cat([average_pred_embeddings, max_pooled_pred_embeddings], -1)
    elif self.config.srl_pred_span_repr == SPAN_REPR_AVE:
      pred_embeddings = self.average_span_extractor(
        sequence_tensor=features, span_indices=predicate_candidates, span_indices_mask=predicate_mask)
    else:
      raise ValueError("The span repr is not defined {}".format(self.config.srl_pred_span_repr))

    if self.config.srl_use_gold_argument:
      num_preds_to_keep = (torch.sum(predicate_mask, dim=-1)).long()
    else:
      num_preds_to_keep = (torch.sum(predicate_mask, dim=-1) * 0.4).long()
      num_preds_to_keep = torch.min(num_preds_to_keep, self.constant_pred)

    (top_pred_span_embeddings, top_pred_span_mask,
     top_pred_span_indices, top_pred_span_mention_scores,
     pred_span_mention_full_scores) = self.predicate_pruner(
      pred_embeddings, predicate_mask, num_preds_to_keep)

    # top_arg_span_mask = top_arg_span_mask.unsqueeze(-1)
    flat_top_arg_span_indices = utils.flatten_and_batch_shift_indices(top_arg_span_indices, num_arg_spans)
    # Shape: (batch, max_arg_to_keep, 2)
    top_arg_spans = utils.batched_index_select(arg_candidates,
                                              top_arg_span_indices,
                                              flat_top_arg_span_indices)

    flat_top_pred_span_indices = utils.flatten_and_batch_shift_indices(
      top_pred_span_indices, num_preds_spans)
    # Shape: (batch, max_pred_to_keep, 2)
    top_pred_spans = utils.batched_index_select(predicate_candidates,
                                                top_pred_span_indices,
                                                flat_top_pred_span_indices)

    # Shape: (batch_size, num_pred_to_keep, num_arg_to_keep, embedding_size)
    span_pair_embeddings = self.compute_span_pair_embeddings(top_pred_span_embeddings,
                                                             top_arg_span_embeddings)

    # Play with the label embeddings
    if "label_candidates" in kwargs and hasattr(self.config, "srl_use_label_embedding") and self.config.srl_use_label_embedding:
      label_candidates = kwargs["label_candidates"]
      label_candidates_mask = kwargs["label_candidates_mask"]

      label_candidate_embeddings_key = self.label_embdding_key(label_candidates)
      label_candidate_embeddings_value = self.label_embdding_value(label_candidates)
      # (batch, max_sent_length (without tokenization), max_label_candidate_size)

      top_pred_label_candidate_embeddings_key = utils.batched_index_select_tensor(label_candidate_embeddings_key,
                                 top_pred_span_indices,
                                 flat_top_pred_span_indices)
      top_pred_label_candidate_embeddings_value = utils.batched_index_select_tensor(label_candidate_embeddings_value,
                                 top_pred_span_indices,
                                 flat_top_pred_span_indices)

      top_pred_label_candidate_mask = utils.batched_index_select_tensor(label_candidates_mask,
                                  top_pred_span_indices,
                                  flat_top_pred_span_indices)
      # (batch, max_pred_candidate_size, max_label_candidate_size, embedding_size)
      max_label_candidate_size = label_candidates.size(2)
      max_argument_size = span_pair_embeddings.size(2)

      transformed_pair_embedding = self.pair_to_embedding(span_pair_embeddings)

      expanded_top_pred_label_candidate_embeddings_key = top_pred_label_candidate_embeddings_key.unsqueeze(2).repeat(
        1, 1, max_argument_size, 1, 1)
      expanded_top_pred_label_candidate_embeddings_value = top_pred_label_candidate_embeddings_value.unsqueeze(2).repeat(
        1, 1, max_argument_size, 1, 1)

      # Attention Method 1
      # expanded_transformed_span_pair_embeddings = transformed_pair_embedding.unsqueeze(3).repeat(
      #   1, 1, 1, max_label_candidate_size, 1)
      # concat = torch.cat([expanded_transformed_span_pair_embeddings, expanded_top_pred_label_candidate_embeddings_key], dim=-1)
      # attn_weights = self.to_attention(concat).squeeze(-1)

      # Attention Method 2
      # (batch_size, predicate_size, argument_size, dim=300)
      # (batch_size, predicate_size, arugment_size, label_size, dim=300)
      attn_weights = torch.matmul(transformed_pair_embedding.unsqueeze(3),
                                  expanded_top_pred_label_candidate_embeddings_key.transpose(-1, -2)).squeeze(3)
      # (batch_size, predicate_size, argument_size, label_size)

      expanded_top_pred_label_candidate_mask = top_pred_label_candidate_mask.unsqueeze(2).repeat(1, 1, max_argument_size, 1)

      attn_weights = utils.masked_softmax(attn_weights, expanded_top_pred_label_candidate_mask)
      # (batch, max_pred_cand, max_arg_cand, max_label_can)
      extra_feature = utils.weighted_sum(expanded_top_pred_label_candidate_embeddings_value, attn_weights)
      span_pair_embeddings = torch.cat([span_pair_embeddings, extra_feature], dim=-1)
      # (batch, max_pred_cand, )

    # if self.config.srl_pair_prune:
    #   matched_predicate_span = kwargs.pop("matched_predicate_span")
    #   # (batch, predicate_spans_num, 2)
    #   predicate_span_to_idx_dict = self.build_span_to_idx_dict(top_pred_spans)
    #   idx_to_collect_from_pair_embedding = []
    #   idx_to_collect_from_example_embedding = []
    #
    #
    #
    #   paired_mask = None
    #   num_pairs_to_keep = None
    #   top_pairs, top_pairs_mask, top_pairs_indices = self.pair_prune(span_pair_embeddings, paired_mask, num_pairs_to_keep)
    #   # top_pairs = (batch, predicate_size, argument_size, embedding_size)
    #   example_pairs = None
    #   # examples = (batch, predicate_size, argument_size', embedding_size)
    #   label_embeddings = None
    #   # label_embeddings = (batch, predicate_size, argument_size', label_embedding)
    #   attention_weights = None
    #   # attention_weights = (batch, predicate_size, argument_size, argument_size')
    #   # label_weighted_sum = attention_weights * label_embeddings = (batch, predicate_size, argument_size, label_embedding)
    #   # zeros = (batch, num_pred_to_keep, num_arg_to_keep, label_embedding_size)
    #   # expanded_label_repr = zeros + label_weighted_sum
    #   expanded_label_repr = None
    #   span_pair_embeddings = torch.cat([span_pair_embeddings, expanded_label_repr], dim=-1)

    """
    Joint Prediction with POS tag and predicate.
    We direct label the head word to see the performance.
    """
    if hasattr(self.config, "srl_compute_pos_tag_loss") and self.config.srl_compute_pos_tag_loss:
      pos_tag_logits = self.to_pos_tag(pos_tag_features)
    else:
      pos_tag_logits = None
    # We directly return to compute the loss

    srl_scores = self.compute_srl_scores(span_pair_embeddings,
                                         top_pred_span_mention_scores,
                                         top_arg_span_mention_scores)

    # Joint POS-Predicate Sequence labeling.

    return (srl_scores, top_pred_spans,
            top_arg_spans, top_pred_span_mask, top_arg_span_mask,
            pred_span_mention_full_scores, arg_span_mention_full_scores, pos_tag_logits)

  def build_span_to_idx_dict(self, spans: torch.LongTensor):
    """
    TODO: Push these operations to C++ if possible
    """
    span_to_idx_dict = {}
    for batch_idx in range(spans.size(0)):
      span_to_idx_dict[batch_idx] = {}
      for idx, span in enumerate(spans[batch_idx]):
        span_data = tuple(span.data.cpu().numpy())
        if span_data not in span_to_idx_dict[batch_idx]:
          span_to_idx_dict[batch_idx][span_data] = idx
    return span_to_idx_dict


  def compute_span_pair_embeddings(self,
                                   top_pred_span_embeddings: torch.FloatTensor,
                                   top_arg_span_embeddings: torch.FloatTensor) -> torch.FloatTensor:
    # Shape (batch_size, num_pred_to_keep, 1, embedding_size)
    num_pred_to_keep = top_pred_span_embeddings.size(1)
    num_arg_to_keep = top_arg_span_embeddings.size(1)
    expanded_pred_embeddings = top_pred_span_embeddings.unsqueeze(2).repeat(1, 1, num_arg_to_keep, 1)
    expanded_arg_embeddings = top_arg_span_embeddings.unsqueeze(1).repeat(1, num_pred_to_keep, 1, 1)
    span_pair_embeddings = torch.cat([expanded_pred_embeddings, expanded_arg_embeddings], dim=-1)
    return span_pair_embeddings

  def compute_srl_scores(self,
                         pairwise_embeddings: torch.FloatTensor,
                         top_pred_span_mention_scores: torch.FloatTensor,
                         top_arg_span_mention_scores: torch.FloatTensor) -> torch.FloatTensor:
    # Shape (batch_size, num_pred_to_keep, num_arg_to_keep, n_classes)
    scores = self.semantic_role_predictor(pairwise_embeddings)
    # top_pred_span_mention_scores = (batch_size, num_pred_to_keep, 1)
    num_pred_to_keep = top_pred_span_mention_scores.size(1)
    num_arg_to_keep = top_arg_span_mention_scores.size(1)
    expanded_top_pred_span_mention_scores = top_pred_span_mention_scores.unsqueeze(2).repeat(1, 1, num_arg_to_keep, 1)
    expanded_top_arg_span_mention_scores = top_arg_span_mention_scores.unsqueeze(1).repeat(1, num_pred_to_keep, 1, 1)
    dummy_scores = torch.zeros(scores.size()[:-1] + (1,)).to(pairwise_embeddings.device)
    scores = scores + expanded_top_arg_span_mention_scores + expanded_top_pred_span_mention_scores
    scores = torch.cat([dummy_scores, scores], dim=-1)
    return scores



from relogic.components.component import Component
from relogic.structures.structure import Structure
from typing import List


class NERComponent(Component):
  """

  """
  def execute(self, inputs: List[Structure]):
    self._predictor(inputs)







import torch.nn as nn
import torch
from relogic.logickit.utils import utils



class AttentiveSpanExtractor(nn.Module):
  def __init__(self, input_dim):
    super(AttentiveSpanExtractor, self).__init__()
    self._global_attention = nn.Linear(input_dim, 1)

  def forward(self,
              sequence_tensor: torch.FloatTensor,
              value_tensor: torch.FloatTensor,
              span_indices: torch.LongTensor,
              sequence_mask: torch.LongTensor = None,
              span_indices_mask: torch.LongTensor = None) -> torch.FloatTensor:
    # Shape (batch_size, num_spans, 1)
    span_starts, span_ends = span_indices.split(1, dim=-1)

    span_ends = span_ends - 1

    span_widths = span_ends - span_starts

    max_batch_span_width = span_widths.max().item() + 1

    global_attention_logits = self._global_attention(sequence_tensor)

    # Shape: (1, 1, max_batch_span_width)
    max_span_range_indices = utils.get_range_vector(max_batch_span_width,
                                                    sequence_tensor.device).view(1, 1, -1)
    span_mask = (max_span_range_indices <= span_widths).float()
    raw_span_indices = span_ends - max_span_range_indices
    span_mask = span_mask * (raw_span_indices >= 0).float()
    span_indices = torch.relu(raw_span_indices.float()).long()

    flat_span_indices = utils.flatten_and_batch_shift_indices(span_indices, sequence_tensor.size(1))

    span_embeddings = utils.batched_index_select(value_tensor, span_indices, flat_span_indices)


    span_attention_logits = utils.batched_index_select(global_attention_logits,
                                                       span_indices,
                                                       flat_span_indices).squeeze(-1)

    span_attention_weights = utils.masked_softmax(span_attention_logits, span_mask)

    attended_text_embeddings = utils.weighted_sum(span_embeddings, span_attention_weights)

    if span_indices_mask is not None:
      return attended_text_embeddings * span_indices_mask.unsqueeze(-1).float()

    return attended_text_embeddings








import torch.nn as nn
import torch
from relogic.logickit.base.utils import log
from typing import Tuple
from relogic.logickit.modules.input_variational_dropout import InputVariationalDropout
from relogic.logickit.modules.bilinear_matrix_attention import BilinearMatrixAttention
import copy
import numpy
from relogic.logickit.utils.utils import get_range_vector, get_device_of, masked_log_softmax
from relogic.logickit.modules.chi_liu_edmonds import decode_mst
import torch.nn.functional as F

class BiaffineDepModule(nn.Module):
  def __init__(self, config, task_name, n_classes):
    super(BiaffineDepModule, self).__init__()
    self.config = config
    self.task_name = task_name
    self.n_classes = n_classes

    if hasattr(self.config, "sequence_labeling_use_cls") and self.config.sequence_labeling_use_cls:
      self.mul = 2
      log("Use CLS in dependency parsing")
    else:
      self.mul = 1

    encoder_dim = config.hidden_size
    arc_representation_dim = tag_representation_dim = config.dep_parsing_mlp_dim
    # self.pos_tag_embedding = nn.Embedding()
    self.head_sentinel = torch.nn.Parameter(torch.randn([1, 1, config.hidden_size]))

    # TODO: Need to check the dropout attribute.
    # TODO: How to design task specific parameter configuration
    self.dropout = InputVariationalDropout(config.dropout)
    self.head_arc_feedforward = nn.Sequential(
      nn.Linear(encoder_dim, arc_representation_dim),
      nn.ELU())
    self.child_arc_feedforward = copy.deepcopy(self.head_arc_feedforward)
    self.head_tag_feedforward = nn.Sequential(
      nn.Linear(encoder_dim, tag_representation_dim),
      nn.ELU())
    self.child_tag_feedforward = copy.deepcopy(self.head_tag_feedforward)

    self.arc_attention = BilinearMatrixAttention(
      matrix_1_dim=arc_representation_dim,
      matrix_2_dim=arc_representation_dim,
      use_input_biases=True)

    self.tag_bilinear = nn.modules.Bilinear(
        tag_representation_dim, tag_representation_dim, self.n_classes)




  def forward(self, *input, **kwargs):
    features = kwargs.pop("features")
    mask = (kwargs.pop("input_head") == 1).long()

    head_indices = kwargs.pop("arcs_ids", None)
    head_tags = kwargs.pop("label_ids", None)


    batch_size = features.size(0)
    encoding_dim = features.size(2)
    head_sentinel = self.head_sentinel.expand(batch_size, 1, encoding_dim)
    encoded_text = torch.cat([head_sentinel, features], dim=1)
    mask = torch.cat([mask.new_ones(batch_size, 1), mask], dim=1)
    float_mask = mask.float()

    if head_indices is not None:
      head_indices = torch.cat([head_indices.new_zeros(batch_size, 1), head_indices], 1)
    if head_tags is not None:
      head_tags = torch.cat([head_tags.new_zeros(batch_size, 1), head_tags], 1)

    encoded_text = self.dropout(encoded_text)

    head_arc_representation = self.dropout(self.head_arc_feedforward(encoded_text))
    child_arc_representation = self.dropout(self.child_arc_feedforward(encoded_text))

    head_tag_representation = self.dropout(self.head_tag_feedforward(encoded_text))
    child_tag_representation = self.dropout(self.child_tag_feedforward(encoded_text))

    attended_arcs = self.arc_attention(head_arc_representation, child_arc_representation)
    # shape (batch_size, sequence_length, sequence_length)

    minus_inf = -1e8
    minus_mask = (1 - float_mask) * minus_inf
    attended_arcs = attended_arcs + minus_mask.unsqueeze(2) + minus_mask.unsqueeze(1)

    if self.training:
      predicted_heads, predicted_head_tags = self.greedy_decode(
        head_tag_representation, child_tag_representation, attended_arcs, mask)

    else:
      predicted_heads, predicted_head_tags = self.mst_decode(
        head_tag_representation, child_tag_representation, attended_arcs, mask
      )

    if head_indices is not None and head_tags is not None:
      arc_nll, tag_nll = self.construct_loss(
        head_tag_representation=head_tag_representation,
        child_tag_representation=child_tag_representation,
        attended_arcs=attended_arcs,
        head_indices=head_indices,
        head_tags=head_tags,
        mask=mask,
      )
    else:
      arc_nll, tag_nll = self.construct_loss(
        head_tag_representation=head_tag_representation,
        child_tag_representation=child_tag_representation,
        attended_arcs=attended_arcs,
        head_indices=predicted_heads.long(),
        head_tags=predicted_head_tags.long(),
        mask=mask,
      )

    loss = arc_nll + tag_nll

    # if head_indices is not None and head_tags is not None:
    #   evaluation_mask = self._get_mask_for_eval(mask[:, 1:], pos_tags)
    #   # We calculate attatchment scores for the whole sentence
    #   # but excluding the symbolic ROOT token at the start,
    #   # which is why we start from the second element in the sequence.
    #   self._attachment_scores(
    #     predicted_heads[:, 1:],
    #     predicted_head_tags[:, 1:],
    #     head_indices,
    #     head_tags,
    #     evaluation_mask,
    #   )

    output_dict = {
      "heads": predicted_heads,
      "head_tags": predicted_head_tags,
      "arc_loss": arc_nll,
      "tag_loss": tag_nll,
      "loss": loss,
      "mask": mask}


    return output_dict



  def construct_loss(
          self,
          head_tag_representation: torch.Tensor,
          child_tag_representation: torch.Tensor,
          attended_arcs: torch.Tensor,
          head_indices: torch.Tensor,
          head_tags: torch.Tensor,
          mask: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Computes the arc and tag loss for a sequence given gold head indices and tags.
    Parameters
    ----------
    head_tag_representation : ``torch.Tensor``, required.
        A tensor of shape (batch_size, sequence_length, tag_representation_dim),
        which will be used to generate predictions for the dependency tags
        for the given arcs.
    child_tag_representation : ``torch.Tensor``, required
        A tensor of shape (batch_size, sequence_length, tag_representation_dim),
        which will be used to generate predictions for the dependency tags
        for the given arcs.
    attended_arcs : ``torch.Tensor``, required.
        A tensor of shape (batch_size, sequence_length, sequence_length) used to generate
        a distribution over attachments of a given word to all other words.
    head_indices : ``torch.Tensor``, required.
        A tensor of shape (batch_size, sequence_length).
        The indices of the heads for every word.
    head_tags : ``torch.Tensor``, required.
        A tensor of shape (batch_size, sequence_length).
        The dependency labels of the heads for every word.
    mask : ``torch.Tensor``, required.
        A mask of shape (batch_size, sequence_length), denoting unpadded
        elements in the sequence.
    Returns
    -------
    arc_nll : ``torch.Tensor``, required.
        The negative log likelihood from the arc loss.
    tag_nll : ``torch.Tensor``, required.
        The negative log likelihood from the arc tag loss.
    """
    float_mask = mask.float()
    batch_size, sequence_length, _ = attended_arcs.size()
    # shape (batch_size, 1)
    range_vector = get_range_vector(batch_size, get_device_of(attended_arcs)).unsqueeze(1)
    # shape (batch_size, sequence_length, sequence_length)
    normalised_arc_logits = (
          masked_log_softmax(attended_arcs, mask)
          * float_mask.unsqueeze(2)
          * float_mask.unsqueeze(1)
    )

    # shape (batch_size, sequence_length, num_head_tags)
    head_tag_logits = self.get_head_tags(
      head_tag_representation, child_tag_representation, head_indices
    )
    normalised_head_tag_logits = masked_log_softmax(
      head_tag_logits, mask.unsqueeze(-1)
    ) * float_mask.unsqueeze(-1)
    # index matrix with shape (batch, sequence_length)
    timestep_index = get_range_vector(sequence_length, get_device_of(attended_arcs))
    child_index = (
      timestep_index.view(1, sequence_length).expand(batch_size, sequence_length).long()
    )
    # shape (batch_size, sequence_length)
    arc_loss = normalised_arc_logits[range_vector, child_index, head_indices]
    tag_loss = normalised_head_tag_logits[range_vector, child_index, head_tags]
    # We don't care about predictions for the symbolic ROOT token's head,
    # so we remove it from the loss.
    arc_loss = arc_loss[:, 1:]
    tag_loss = tag_loss[:, 1:]

    # The number of valid positions is equal to the number of unmasked elements minus
    # 1 per sequence in the batch, to account for the symbolic HEAD token.
    valid_positions = mask.sum() - batch_size

    arc_nll = -arc_loss.sum() / valid_positions.float()
    tag_nll = -tag_loss.sum() / valid_positions.float()
    return arc_nll, tag_nll

  def greedy_decode(
        self,
        head_tag_representation: torch.Tensor,
        child_tag_representation: torch.Tensor,
        attended_arcs: torch.Tensor,
        mask: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Decodes the head and head tag predictions by decoding the unlabeled arcs
    independently for each word and then again, predicting the head tags of
    these greedily chosen arcs independently. Note that this method of decoding
    is not guaranteed to produce trees (i.e. there maybe be multiple roots,
    or cycles when children are attached to their parents).
    Parameters
    ----------
    head_tag_representation : ``torch.Tensor``, required.
        A tensor of shape (batch_size, sequence_length, tag_representation_dim),
        which will be used to generate predictions for the dependency tags
        for the given arcs.
    child_tag_representation : ``torch.Tensor``, required
        A tensor of shape (batch_size, sequence_length, tag_representation_dim),
        which will be used to generate predictions for the dependency tags
        for the given arcs.
    attended_arcs : ``torch.Tensor``, required.
        A tensor of shape (batch_size, sequence_length, sequence_length) used to generate
        a distribution over attachments of a given word to all other words.
    Returns
    -------
    heads : ``torch.Tensor``
        A tensor of shape (batch_size, sequence_length) representing the
        greedily decoded heads of each word.
    head_tags : ``torch.Tensor``
        A tensor of shape (batch_size, sequence_length) representing the
        dependency tags of the greedily decoded heads of each word.
    """
    attended_arcs = attended_arcs + torch.diag(
      attended_arcs.new(mask.size(1)).fill_(-numpy.inf))
    # Compute the heads greedily.
    # shape (batch_size, sequence_length)
    _, heads = attended_arcs.max(dim=2)
    # Given the greedily predicted heads, decode their dependency tags.
    # shape (batch_size, sequence_length, num_head_tags)
    head_tag_logits = self.get_head_tags(
      head_tag_representation, child_tag_representation, heads
    )
    _, head_tags = head_tag_logits.max(dim=2)
    return heads, head_tags

  def get_head_tags(
        self,
        head_tag_representation: torch.Tensor,
        child_tag_representation: torch.Tensor,
        head_indices: torch.Tensor) -> torch.Tensor:
    """
    Decodes the head tags given the head and child tag representations
    and a tensor of head indices to compute tags for. Note that these are
    either gold or predicted heads, depending on whether this function is
    being called to compute the loss, or if it's being called during inference.
    Parameters
    ----------
    head_tag_representation : ``torch.Tensor``, required.
        A tensor of shape (batch_size, sequence_length, tag_representation_dim),
        which will be used to generate predictions for the dependency tags
        for the given arcs.
    child_tag_representation : ``torch.Tensor``, required
        A tensor of shape (batch_size, sequence_length, tag_representation_dim),
        which will be used to generate predictions for the dependency tags
        for the given arcs.
    head_indices : ``torch.Tensor``, required.
        A tensor of shape (batch_size, sequence_length). The indices of the heads
        for every word.
    Returns
    -------
    head_tag_logits : ``torch.Tensor``
        A tensor of shape (batch_size, sequence_length, num_head_tags),
        representing logits for predicting a distribution over tags
        for each arc.
    """
    batch_size = head_tag_representation.size(0)
    # shape (batch_size,)
    range_vector = get_range_vector(
      batch_size, get_device_of(head_tag_representation)
    ).unsqueeze(1)

    # This next statement is quite a complex piece of indexing, which you really
    # need to read the docs to understand. See here:
    # https://docs.scipy.org/doc/numpy-1.13.0/reference/arrays.indexing.html#advanced-indexing
    # In effect, we are selecting the indices corresponding to the heads of each word from the
    # sequence length dimension for each element in the batch.

    # shape (batch_size, sequence_length, tag_representation_dim)
    selected_head_tag_representations = head_tag_representation[range_vector, head_indices]
    selected_head_tag_representations = selected_head_tag_representations.contiguous()
    # shape (batch_size, sequence_length, num_head_tags)
    head_tag_logits = self.tag_bilinear(
      selected_head_tag_representations, child_tag_representation
    )
    return head_tag_logits

  def mst_decode(
        self,
        head_tag_representation: torch.Tensor,
        child_tag_representation: torch.Tensor,
        attended_arcs: torch.Tensor,
        mask: torch.Tensor,
  ) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Decodes the head and head tag predictions using the Edmonds' Algorithm
    for finding minimum spanning trees on directed graphs. Nodes in the
    graph are the words in the sentence, and between each pair of nodes,
    there is an edge in each direction, where the weight of the edge corresponds
    to the most likely dependency label probability for that arc. The MST is
    then generated from this directed graph.
    Parameters
    ----------
    head_tag_representation : ``torch.Tensor``, required.
        A tensor of shape (batch_size, sequence_length, tag_representation_dim),
        which will be used to generate predictions for the dependency tags
        for the given arcs.
    child_tag_representation : ``torch.Tensor``, required
        A tensor of shape (batch_size, sequence_length, tag_representation_dim),
        which will be used to generate predictions for the dependency tags
        for the given arcs.
    attended_arcs : ``torch.Tensor``, required.
        A tensor of shape (batch_size, sequence_length, sequence_length) used to generate
        a distribution over attachments of a given word to all other words.
    Returns
    -------
    heads : ``torch.Tensor``
        A tensor of shape (batch_size, sequence_length) representing the
        greedily decoded heads of each word.
    head_tags : ``torch.Tensor``
        A tensor of shape (batch_size, sequence_length) representing the
        dependency tags of the optimally decoded heads of each word.
    """
    batch_size, sequence_length, tag_representation_dim = head_tag_representation.size()

    lengths = mask.data.sum(dim=1).long().cpu().numpy()

    expanded_shape = [batch_size, sequence_length, sequence_length, tag_representation_dim]
    head_tag_representation = head_tag_representation.unsqueeze(2)
    head_tag_representation = head_tag_representation.expand(*expanded_shape).contiguous()
    child_tag_representation = child_tag_representation.unsqueeze(1)
    child_tag_representation = child_tag_representation.expand(*expanded_shape).contiguous()
    # Shape (batch_size, sequence_length, sequence_length, num_head_tags)
    pairwise_head_logits = self.tag_bilinear(head_tag_representation, child_tag_representation)

    # Note that this log_softmax is over the tag dimension, and we don't consider pairs
    # of tags which are invalid (e.g are a pair which includes a padded element) anyway below.
    # Shape (batch, num_labels,sequence_length, sequence_length)
    normalized_pairwise_head_logits = F.log_softmax(pairwise_head_logits, dim=3).permute(
      0, 3, 1, 2
    )

    # Mask padded tokens, because we only want to consider actual words as heads.
    minus_inf = -1e8
    minus_mask = (1 - mask.float()) * minus_inf
    attended_arcs = attended_arcs + minus_mask.unsqueeze(2) + minus_mask.unsqueeze(1)

    # Shape (batch_size, sequence_length, sequence_length)
    normalized_arc_logits = F.log_softmax(attended_arcs, dim=2).transpose(1, 2)

    # Shape (batch_size, num_head_tags, sequence_length, sequence_length)
    # This energy tensor expresses the following relation:
    # energy[i,j] = "Score that i is the head of j". In this
    # case, we have heads pointing to their children.
    batch_energy = torch.exp(
      normalized_arc_logits.unsqueeze(1) + normalized_pairwise_head_logits
    )
    return self._run_mst_decoding(batch_energy, lengths)

  @staticmethod
  def _run_mst_decoding(
        batch_energy: torch.Tensor, lengths: torch.Tensor
  ) -> Tuple[torch.Tensor, torch.Tensor]:
    heads = []
    head_tags = []
    for energy, length in zip(batch_energy.detach().cpu(), lengths):
      scores, tag_ids = energy.max(dim=0)
      # Although we need to include the root node so that the MST includes it,
      # we do not want any word to be the parent of the root node.
      # Here, we enforce this by setting the scores for all word -> ROOT edges
      # edges to be 0.
      scores[0, :] = 0
      # Decode the heads. Because we modify the scores to prevent
      # adding in word -> ROOT edges, we need to find the labels ourselves.
      instance_heads, _ = decode_mst(scores.numpy(), length, has_labels=False)

      # Find the labels which correspond to the edges in the max spanning tree.
      instance_head_tags = []
      for child, parent in enumerate(instance_heads):
        instance_head_tags.append(tag_ids[parent, child].item())
      # We don't care what the head or tag is for the root token, but by default it's
      # not necesarily the same in the batched vs unbatched case, which is annoying.
      # Here we'll just set them to zero.
      instance_heads[0] = 0
      instance_head_tags[0] = 0
      heads.append(instance_heads)
      head_tags.append(instance_head_tags)
    return torch.from_numpy(numpy.stack(heads)), torch.from_numpy(numpy.stack(head_tags))

  def _get_mask_for_eval(
        self, mask: torch.LongTensor, pos_tags: torch.LongTensor
  ) -> torch.LongTensor:
    """
    Dependency evaluation excludes words are punctuation.
    Here, we create a new mask to exclude word indices which
    have a "punctuation-like" part of speech tag.
    Parameters
    ----------
    mask : ``torch.LongTensor``, required.
        The original mask.
    pos_tags : ``torch.LongTensor``, required.
        The pos tags for the sequence.
    Returns
    -------
    A new mask, where any indices equal to labels
    we should be ignoring are masked.
    """
    new_mask = mask.detach()
    for label in self._pos_to_ignore:
      label_mask = pos_tags.eq(label).long()
      new_mask = new_mask * (1 - label_mask)
    return new_mask










from relogic.logickit.base.constants import *
from relogic.logickit.tasks.task import Task

import torch.nn.functional as F
import torch

def get_loss(task: Task, logits, label_ids, config, extra_arg=dict({}), input_head = None, **kwargs):
  if task.name.startswith(IR_TASK):
    if label_ids is None:
      label_ids = kwargs.pop("_label_ids")
    if config.regression:
      return F.binary_cross_entropy_with_logits(logits.squeeze(1), label_ids)
    else:
      return F.cross_entropy(logits, label_ids)
  if task.name.startswith(DOCIR_TASK):
    if config.regression:
      return F.binary_cross_entropy_with_logits(logits.squeeze(1), label_ids[0].unsqueeze(0))
    else:
      return F.cross_entropy(logits, label_ids[0].unsqueeze(0))
  if task.name in ["joint_srl"]:
    if isinstance(label_ids, tuple):
      label_ids, pred_span_label, arg_span_label, pos_tag_ids = label_ids
    batch_size = label_ids.size(0)
    key = label_ids[:, :, :4]
    batch_id = torch.arange(0, batch_size).unsqueeze(1).unsqueeze(1).repeat(1, key.size(1), 1).to(key.device)
    expanded_key = torch.cat([batch_id, key], dim=-1)
    v = label_ids[:, :, 4]
    # batch_id = torch.arange(0, v.size(0)).unsqueeze(1).repeat(1, v.size(1)).to(key.device)
    # expanded_v = torch.cat([batch_id.unsqueeze(-1), v.unsqueeze(-1)], dim=-1)
    flatten_key = expanded_key.view(-1, 5)
    flatten_v = v.view(-1)

    (srl_scores, top_pred_spans, top_arg_spans, top_pred_span_mask, top_arg_span_mask,
     pred_span_mention_full_scores, arg_span_mention_full_scores, pos_tag_logits) = logits
    # (batch_size, max_pred_num, 2), (batch_size, max_arg_num, 2)
    max_pred_num = top_pred_spans.size(1)
    max_arg_num = top_arg_spans.size(1)
    expanded_top_pred_spans = top_pred_spans.unsqueeze(2).repeat(1, 1, max_arg_num, 1)
    expanded_top_arg_spans = top_arg_spans.unsqueeze(1).repeat(1, max_pred_num, 1, 1)
    indices = torch.cat([expanded_top_pred_spans, expanded_top_arg_spans], dim=-1)
    batch_id = torch.arange(0, batch_size).unsqueeze(1).unsqueeze(1).unsqueeze(1).repeat(1, *indices.size()[1:3], 1).to(
      indices.device)
    expanded_indices = torch.cat([batch_id, indices], dim=-1)
    flatten_expanded_indices = expanded_indices.view(-1, 5)

    # Generate Loss Mask
    flatten_expanded_top_pred_span_mask = top_pred_span_mask.unsqueeze(2).repeat(1, 1, max_arg_num).view(-1)
    # (batch_size, max_pred_to_keep)
    flatten_expanded_top_arg_span_mask = top_arg_span_mask.unsqueeze(2).repeat(1, max_pred_num, 1).view(-1)
    # (batch_size, max_arg_to_keep)
    merged_mask = flatten_expanded_top_pred_span_mask & flatten_expanded_top_arg_span_mask

    # build dictionary
    d = {}
    for key, value in zip(flatten_key.cpu().numpy(), flatten_v.cpu().numpy()):
      d[tuple(key)] = value

    label_list = []
    for index in flatten_expanded_indices.cpu().numpy():
      label_list.append(d.get(tuple(index), 0))

    # arg_boundary = max(torch.max(top_arg_spans).item(), torch.max(key[:,:,2:]).item()) + 1
    # pred_boundary= max(torch.max(top_pred_spans).item(), torch.max(key[:,:,:2]).item()) + 1
    # size = (batch_size, pred_boundary, pred_boundary, arg_boundary, arg_boundary)
    #
    # dense_label = torch.sparse.LongTensor(flatten_key.t(), flatten_v, size).to(key.device)
    #
    # selected_label = dense_label.masked_select(expanded_indices)

    selected_label = torch.LongTensor(label_list).to(label_ids.device)

    label_loss = F.cross_entropy(srl_scores.view(-1, srl_scores.size(-1))[merged_mask == 1], selected_label[merged_mask == 1])

    # Compute the unary scorer loss
    if hasattr(task.config, "srl_candidate_loss") and task.config.srl_candidate_loss:
      flatten_pred_span_mention_full_scores = F.sigmoid(pred_span_mention_full_scores).view(-1)
      flatten_arg_span_mention_full_scores = F.sigmoid(arg_span_mention_full_scores).view(-1)
      flatten_pred_span_label = pred_span_label.view(-1).float()
      flatten_arg_span_label = arg_span_label.view(-1).float()
      srl_pred_candidate_loss = F.binary_cross_entropy(flatten_pred_span_mention_full_scores, flatten_pred_span_label)
      srl_arg_candidate_loss = F.binary_cross_entropy(flatten_arg_span_mention_full_scores, flatten_arg_span_label)
      candidate_loss = srl_pred_candidate_loss + srl_arg_candidate_loss
      return candidate_loss + label_loss

    if hasattr(task.config, "srl_compute_pos_tag_loss") and task.config.srl_compute_pos_tag_loss:
      active_loss = input_head[:, :pos_tag_logits.size(1)].contiguous().view(-1) == 1
      # I use pos_tag_logits.size(1) to get the label length. It is OK to filter extra things
      active_pos_tag_logits = pos_tag_logits.view(-1, pos_tag_logits.size(-1))[active_loss]
      active_labels = pos_tag_ids[:, :pos_tag_logits.size(1)].contiguous().view(-1)[active_loss]
      loss = F.cross_entropy(active_pos_tag_logits, active_labels)
      label_loss += loss

    return label_loss
  elif task.name in [NER_TASK, POS_TASK, PIPE_SRL_TASK, PREDICATE_DETECTION_TASK]:
    if input_head is not None:
      # Use the BERT based model
      active_loss = input_head[:, :logits.size(1)].contiguous().view(-1) == 1
      # I use logits.size(1) to get the label length. It is OK to filter extra things
    else:
      # create a mask based on the sequence length
      active_loss = kwargs.pop("_input_token_mask").view(-1)
      # We need to assign the value of _label_ids to label_ids
      label_ids = kwargs.pop("_label_ids")
    active_logits = logits.view(-1, logits.size(-1))[active_loss]
    active_labels = label_ids[:, :logits.size(1)].contiguous().view(-1)[active_loss]
    loss = F.cross_entropy(active_logits, active_labels)

    return loss
  elif task.name in [ENTITY_TYPE_CLASSIFICATION]:
    loss = F.binary_cross_entropy_with_logits(logits, label_ids.float())
    return loss

  elif task.name in [PARALLEL_TEACHER_STUDENT_TASK]:
    active_loss = kwargs.pop("mask")
    if active_loss is not None:
      active_loss = kwargs.pop("mask").view(-1)
      target = kwargs.pop("target")
      active_logits = F.softmax(logits.view(-1, logits.size(-1)), -1)[active_loss]
      active_target = F.softmax(target.view(-1, target.size(-1)), -1)[active_loss]
      loss = F.kl_div(active_logits.log(), active_target)
    else:
      target = kwargs.pop("target")
      if task.config.use_cosine_loss:
        loss = (1 - F.cosine_similarity(logits, target)).sum(-1) / logits.size(0)
      else:
        loss = F.l1_loss(logits, target)
    return loss
  elif task.name in [MIXSENT_TASK]:
    target = kwargs.pop("target")
    # loss = F.mse_loss(logits, target)
    active_logits = F.softmax(logits.view(-1, logits.size(-1)), -1)
    active_target = F.softmax(target.view(-1, target.size(-1)), -1)
    loss = F.kl_div(active_logits.log(), active_target, reduction="batchmean")
    return loss
  else:
    span_boundary, logits = logits
    return F.cross_entropy(logits.view(-1, logits.size(-1)), label_ids.view(-1))

# def get_loss(task_name, logits, label_ids, config, extra_args):
#   if task_name in AUTO_SPAN:
#     """
#     For auto span task, there are two parts in logits:
#       one is for span detection
#       the other one is for label prediction
#     Basically span detection is a sequence labeling task.
#       We do not only consider the head of token. So it is
#       full version of sequence labeling
#     """
#     span_logits, label_logits = logits
#     token_label_ids, span_label_ids = label_ids
#     assert "segment_ids" in extra_args, "segment_ids should be in extra_args"
#     assert "input_mask" in extra_args, "input_mask should be in extra_args"
#     # We use segment ids to get sentence labeling
#     segment_ids = extra_args["segment_ids"]
#     input_mask = extra_args["input_mask"]
#     active_loss = (segment_ids[:, :span_logits.size(1)].contiguous().view(-1) == 0) & \
#                   (input_mask[:, :span_logits.size(1)].contiguous().view(-1) == 1)
#     # This operation ensure that it is first part of the input
#     active_logits = span_logits.view(-1, span_logits.size(-1))[active_loss]
#     active_labels = token_label_ids[:, :span_logits.size(1)].contiguous().view(-1)[active_loss]
#     token_loss = F.cross_entropy(active_logits, active_labels)
#
#     assert "span_mask" in extra_args, "span_mask should be in extra_args"
#     span_mask = extra_args["span_mask"]
#     active_loss = span_mask.view(-1) == 1
#     active_logits = label_logits.view(-1, label_logits.size(-1))[active_loss]
#     active_labels = span_label_ids.view(-1)[active_loss]
#     span_loss = F.cross_entropy(active_logits, active_labels)
#
#     return token_loss + span_loss






from typing import Tuple, Union

import torch
import torch.nn as nn

from relogic.logickit.utils import utils

class Pruner(nn.Module):
  """
  This module scores and prunes items in a list using a parameterized scoring function
  and a threshold
  """
  def __init__(self, scorer: nn.Module) -> None:
    super(Pruner, self).__init__()
    self.scorer = scorer

  def forward(self,
              embeddings: torch.FloatTensor,
              mask: torch.LongTensor,
              num_items_to_keep: Union[int, torch.LongTensor]) -> Tuple[torch.FloatTensor,
      torch.LongTensor, torch.LongTensor, torch.FloatTensor, torch.FloatTensor]:
    """

    Args:
      embeddings: A tensor of shape (batch_size, num_items, embedding_size),
        containing an embedding for each item in the list that we want to prune.
      mask: A tensor of shape (batch_size, num_items), denoting unpadded elements
        of ``embeddings``.
      num_items_to_keep: If a tensor of shape (batch_size), specifies the number of
        items to keep for each individual sentence in minibatch.
        If an int, keep the same number of items for all sentences.

    """
    batch_size = mask.size(0)
    if isinstance(num_items_to_keep, int):
      num_items_to_keep = num_items_to_keep * torch.ones([batch_size],
                                                         dtype=torch.long, device=mask.device)
    max_items_to_keep = num_items_to_keep.max()

    # Shape (batch_size, num_items, 1)
    mask = mask.unsqueeze(-1)

    num_items = embeddings.size(1)
    # Shape (batch_size, num_items, 1)
    scores = self.scorer(embeddings)

    if scores.size(-1) != 1 or scores.dim() != 3:
      raise ValueError("The scorer passed to Pruner must produce a tensor of "
                       "shape ({}, {}, 1), but found shpae {}".format(batch_size, num_items, scores.size()))

    scores = utils.replace_masked_values(scores, mask, -1e20)
    # can not support k as tensor. So for each sentence, we keep the same size k which is max.
    _, top_indices = scores.topk(max_items_to_keep, 1)

    # Mask based on the number of items to keep for each sentence
    top_indices_mask = utils.get_mask_from_sequence_lengths(num_items_to_keep, max_items_to_keep)
    top_indices_mask = top_indices_mask.byte()

    # Shape (batch, max_items_to_keep)
    top_indices = top_indices.squeeze(-1)

    fill_value, _ = top_indices.max(dim=1)
    fill_value = fill_value.unsqueeze(-1)

    top_indices = torch.where(top_indices_mask, top_indices, fill_value)

    top_indices, _ = torch.sort(top_indices, 1)

    flat_top_indices = utils.flatten_and_batch_shift_indices(top_indices, num_items)

    top_embeddings = utils.batched_index_select(embeddings, top_indices, flat_top_indices)

    sequence_mask = utils.batched_index_select(mask, top_indices, flat_top_indices)
    sequence_mask = sequence_mask.squeeze(-1).byte()
    top_mask = top_indices_mask & sequence_mask
    top_mask = top_mask.long()

    # Shape: (batch_size, max_num_items_to_keep, 1)
    top_scores = utils.batched_index_select(scores, top_indices, flat_top_indices)

    return top_embeddings, top_mask, top_indices, top_scores, scores







import torch.nn as nn
import torch
from relogic.logickit.base.utils import log

class SequenceLabelingModule(nn.Module):
  def __init__(self, config, task_name, n_classes):
    super(SequenceLabelingModule, self).__init__()
    self.config = config
    self.task_name = task_name
    self.n_classes = n_classes
    if hasattr(self.config, "sequence_labeling_use_cls") and self.config.sequence_labeling_use_cls:
      self.mul = 2
      log("Use cls in sequence labeling")
    else:
      self.mul = 1
    self.projection = nn.Linear(config.hidden_size * self.mul, config.projection_size)
    self.to_logits = nn.Linear(config.projection_size, self.n_classes)
    self.init_weight()

  def init_weight(self):
    self.to_logits.bias.data.zero_()
    self.to_logits.weight.data.normal_(mean=0.0, std=self.config.initializer_range)

  def forward(self, *input, **kwargs):
    features = kwargs.pop("features")
    if self.mul == 2:
      features = torch.cat([features, features[:, 0].unsqueeze(1).repeat(1, features.size(1), 1)], dim=-1)
    logits = self.to_logits(self.projection(features))
    return logits